"""migrations command group"""

from typing import TYPE_CHECKING

import click
from click import Argument, Context
from click.shell_completion import CompletionItem

from cmem_cmemc.command import CmemcCommand
from cmem_cmemc.command_group import CmemcGroup
from cmem_cmemc.completion import check_option_in_params, finalize_completion
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.migrations.access_conditions_243 import (
    MoveAccessConditionsToNewGraph,
    RenameAuthVocabularyResources,
)
from cmem_cmemc.migrations.bootstrap_data import MigrateBootstrapData
from cmem_cmemc.migrations.remove_noop_triple_251 import RemoveHideHeaderFooterStatements
from cmem_cmemc.migrations.shapes_widget_integrations_243 import (
    ChartsOnNodeShapesToToWidgetIntegrations,
    ChartsOnPropertyShapesToWidgetIntegrations,
    TableReportPropertyShapesToWidgetIntegrations,
    WorkflowTriggerPropertyShapesToWidgetIntegrations,
)
from cmem_cmemc.migrations.sparql_query_texts_242 import SparqlDatatypesToXsdString
from cmem_cmemc.migrations.workspace_configurations import MigrateWorkspaceConfiguration
from cmem_cmemc.object_list import (
    DirectListPropertyFilter,
    DirectValuePropertyFilter,
    ObjectList,
    compare_regex,
    transform_lower,
)

if TYPE_CHECKING:
    from cmem_cmemc.migrations.abc import MigrationRecipe


def get_migrations(ctx: click.Context) -> list[dict]:  # noqa: ARG001
    """Get users for object list"""
    data = [
        {
            "id": _.id,
            "description": _.description,
            "component": _.component,
            "first_version": _.first_version,
            "last_version": _.last_version,
            "tags": _.tags,
            "object": _,
        }
        for _ in [
            MoveAccessConditionsToNewGraph(),
            RenameAuthVocabularyResources(),
            MigrateBootstrapData(),
            MigrateWorkspaceConfiguration(),
            ChartsOnNodeShapesToToWidgetIntegrations(),
            ChartsOnPropertyShapesToWidgetIntegrations(),
            WorkflowTriggerPropertyShapesToWidgetIntegrations(),
            TableReportPropertyShapesToWidgetIntegrations(),
            SparqlDatatypesToXsdString(),
            RemoveHideHeaderFooterStatements(),
        ]
    ]
    data.sort(key=lambda x: x["first_version"])
    return data


def complete_migration_ids(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of migration recipe IDs"""
    ApplicationContext.set_connection_from_params(ctx.find_root().params)
    options = []
    for _ in get_migrations(ctx=ctx):
        id_ = _["id"]
        description = _["description"]
        if check_option_in_params(id_, ctx.params.get(param.name)):  # type: ignore[attr-defined, arg-type]
            continue
        options.append((id_, description))
    return finalize_completion(candidates=options, incomplete=incomplete)


migrations_list = ObjectList(
    name="migration recipes",
    get_objects=get_migrations,
    filters=[
        DirectValuePropertyFilter(
            name="id",
            description="Filter migrations by id.",
            property_key="id",
            transform=transform_lower,
        ),
        DirectValuePropertyFilter(
            name="description",
            description="Filter migrations by regex over description.",
            property_key="description",
            compare=compare_regex,
            fixed_completion=[],
        ),
        DirectValuePropertyFilter(
            name="first_version",
            description="Filter migrations by first version which is target of this migration.",
            property_key="first_version",
            transform=transform_lower,
        ),
        DirectListPropertyFilter(
            name="tag",
            description="Filter migrations by tags",
            property_key="tags",
        ),
    ],
)


@click.command(cls=CmemcCommand, name="list")
@click.option(
    "--filter",
    "filter_",
    type=(str, str),
    multiple=True,
    help=migrations_list.get_filter_help_text(),
    shell_complete=migrations_list.complete_values,
)
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only IDs. This is useful for piping the IDs into other commands.",
)
@click.option("--raw", is_flag=True, help="Outputs raw JSON.")
@click.pass_context
def list_command(
    ctx: click.Context, filter_: tuple[tuple[str, str]], id_only: bool, raw: bool
) -> None:
    """List migration recipies.

    This command lists all available migration recipies
    """
    app: ApplicationContext = ctx.obj
    data = migrations_list.apply_filters(ctx=ctx, filter_=filter_)
    if raw:
        # https://stackoverflow.com/questions/17665809/
        app.echo_info_json([(_, _.pop("object"))[0] for _ in data])
        return
    if id_only:
        for _ in sorted([_.get("id") for _ in data]):
            app.echo_info(_)
        return

    table = [
        [
            _.get("id"),
            _.get("description"),
            ", ".join(sorted(_.get("tags"))),
            f"{_.get('first_version')} ({_.get('component')})",
        ]
        for _ in data
    ]
    app.echo_info_table(
        table,
        headers=["ID", "Description", "Tags", "First Version"],
        sort_column=3,
        caption=f"{len(table)} migration(s)",
        empty_table_message="No migrations available.",
    )


@click.command(cls=CmemcCommand, name="execute")
@click.argument(
    "migration_id", type=click.STRING, required=False, shell_complete=complete_migration_ids
)
@click.option(
    "--filter",
    "filter_",
    type=(str, str),
    multiple=True,
    help=migrations_list.get_filter_help_text(),
    shell_complete=migrations_list.complete_values,
)
@click.option(
    "-a",
    "--all",
    "all_",
    is_flag=True,
    help="Execute all needed migrations.",
)
@click.option("--test-only", is_flag=True, help="Only test, do not execute migrations.")
@click.option("--id-only", is_flag=True, help="Lists only recipe identifier. ")
@click.pass_context
def execute_command(  # noqa: PLR0913
    ctx: click.Context,
    migration_id: str,
    filter_: tuple[tuple[str, str]],
    all_: bool,
    test_only: bool,
    id_only: bool,
) -> None:
    """Execute needed migration recipes.

    This command executes one or more migration recipes.
    Each recipe has a check method to determine if a migration is needed.
    In addition to that, the current component version needs to match the specified
    first-last-version range of the recipe.

    Recipes are executed ordered by first_version.

    Here are some argument examples, in order to see how to use this command:
    execute --all --test-only will list all needed migrations (but not execute them),
    execute --filter tag system will apply all migrations which target system data,
    execute bootstrap-data will apply bootstrap-data migration if needed.
    """
    app: ApplicationContext = ctx.obj
    if not all_ and not migration_id and not filter_:
        raise click.UsageError(
            "You can execute a single recipe, some recipes based on filter, or all recipes. "
            "See the documentation for more information."
        )
    data = migrations_list.apply_filters(ctx=ctx, filter_=filter_)
    if migration_id:
        data = migrations_list.apply_filters(ctx=ctx, filter_=[("id", migration_id)])
        if not data:
            raise click.UsageError(
                f"Migration recipe '{migration_id}' not found. "
                "Use the 'migration list' command to get available migration recipes."
            )
    applied_counter = 0
    for _ in data:
        recipe: MigrationRecipe = _.get("object")
        if not recipe.version_matches():
            app.echo_debug(f"Migration '{recipe.id}' does not match component version.")
            continue
        if not recipe.is_applicable():
            app.echo_debug(f"Migration '{recipe.id}' is not applicable.")
            continue
        app.echo_debug(f"Migration '{recipe.id}' could be applied.")
        applied_counter += 1

        app.echo_info(recipe.id, condition=id_only)
        app.echo_info(f"{recipe.description} ({recipe.id}) ... ", nl=False, condition=not id_only)
        if test_only:
            app.echo_warning("needed", condition=not id_only)
            continue
        recipe.apply()
        app.echo_success("done", condition=not id_only)
    app.echo_success("No migration needed.", condition=(not applied_counter and not id_only))


@click.group(cls=CmemcGroup)
def migration() -> CmemcGroup:  # type: ignore[empty-body]
    """List and apply migration recipes.

    With this command group, you can check your instance for needed
    migration activities as well as apply them to your data.

    Beside an ID and a description, migration recipes have the following
    metadata: 'First Version' is the first Corporate Memory version,
    where this recipe is maybe applicable. The recipe will never be applied
    to a version below this version. 'Tags' is a classification of the recipe
    with regard to the target data, it migrates.

    The following tags are important:
    `system` recipes target data structures
    which are needed to run the most basic functionality properly. These recipes
    can and should be applied after each version upgrade.
    `user` recipes can change user and / or customizing data.
    `acl` recipes migrate access condition data.
    `shapes` recipes migrate shape data.
    `config` recipes migrate configuration data.
    """


migration.add_command(list_command)
migration.add_command(execute_command)



"""manual command for cmem command line interface"""

import click
from click import Context, UsageError

from cmem_cmemc.command import CmemcCommand
from cmem_cmemc.manual_helper.graph import print_manual_graph
from cmem_cmemc.manual_helper.multi_page import create_multi_page_documentation
from cmem_cmemc.manual_helper.single_page import print_manual
from cmem_cmemc.parameter_types.path import ClickSmartPath
from cmem_cmemc.utils import get_version

FORMAT_MD_FILE = "markdown-single-page"
FORMAT_MD_DIRECTORY = "markdown-multi-page"
FORMAT_TURTLE_FILE = "turtle"


@click.command(cls=CmemcCommand, name="manual", hidden=True)
@click.option(
    "--format",
    "format_",
    type=click.Choice([FORMAT_MD_FILE, FORMAT_MD_DIRECTORY, FORMAT_TURTLE_FILE]),
    help=f"Output format. The '{FORMAT_TURTLE_FILE}' and '{FORMAT_MD_FILE}' formats will "
    f"be returned to stdout. The '{FORMAT_MD_DIRECTORY}' format creates a directory tree "
    f"(use '--output-dir' to specify the root directory)",
)
@click.option(
    "--output-dir",
    type=ClickSmartPath(writable=True, file_okay=False),
    help=f"The output directory to create the '{FORMAT_MD_DIRECTORY}' documentation. "
    f"Warning: Existing directories will be overwritten.",
)
@click.pass_context
def manual_command(ctx: Context, format_: str, output_dir: str) -> None:
    """Generate reference documentation assets.

    This command generates the cmemc reference documentation from the help texts
    and outputs it in different formats.
    """
    if format_ is None:
        raise UsageError(
            "Please use the --format option to specify an output format. "
            "Use --help for more information."
        )
    if format_ == FORMAT_TURTLE_FILE:
        print_manual_graph(ctx.find_root(), get_version())
        return
    if format_ == FORMAT_MD_FILE:
        print_manual(ctx.find_root())
        return
    if output_dir is None:
        raise UsageError(
            f"The output format '{FORMAT_MD_DIRECTORY}' needs an output directory "
            f"('--output-dir'). Use --help for more information."
        )
    create_multi_page_documentation(ctx.find_root(), str(output_dir))



"""configuration commands for cmem command line interface."""

import click
from click import ClickException

from cmem_cmemc.command import CmemcCommand
from cmem_cmemc.command_group import CmemcGroup
from cmem_cmemc.context import KNOWN_CONFIG_KEYS, ApplicationContext


@click.command(cls=CmemcCommand, name="list")
@click.pass_obj
def list_command(app: ApplicationContext) -> None:
    """List configured connections.

    This command lists all configured
    connections from the currently used config file.

    The connection identifier can be used with the --connection option
    in order to use a specific Corporate Memory instance.

    In order to apply commands on more than one instance, you need to use
    typical unix gear such as xargs or parallel.

    Example: cmemc config list | xargs -I % sh -c 'cmemc -c % admin status'

    Example: cmemc config list | parallel --jobs 5 cmemc -c {} admin status
    """
    for section_string in sorted(app.get_config(), key=str.casefold):
        if section_string != "DEFAULT":
            app.echo_result(section_string)


@click.command(cls=CmemcCommand, name="edit")
@click.pass_obj
def edit_command(app: ApplicationContext) -> None:
    """Edit the user-scope configuration file."""
    app.echo_info("Open editor for config file " + str(app.config_file))
    click.edit(filename=str(app.config_file))


@click.command(cls=CmemcCommand, name="get")
@click.argument(
    "KEY", nargs=1, type=click.Choice(list(KNOWN_CONFIG_KEYS.keys()), case_sensitive=False)
)
@click.pass_obj
def get_command(app: ApplicationContext, key: str) -> None:
    """Get the value of a known cmemc configuration key.

    In order to automate processes such as fetching custom API data
    from multiple Corporate Memory instances, this command provides a way to
    get the value of a cmemc configuration key for the selected deployment.

    Example: curl -H "Authorization: Bearer $(cmemc -c my admin token)"
    $(cmemc -c my config get DP_API_ENDPOINT)/api/custom/slug

    The commands return with exit code 1 if the config key is not used in
    the current configuration.
    """
    value = KNOWN_CONFIG_KEYS[key]()
    app.echo_debug(f"Type of {key} value is {type(value)}")
    if value is None:
        raise ClickException(f"Configuration key {key} is not used in this configuration.")
    app.echo_info(str(value))


@click.command(cls=CmemcCommand, name="eval")
@click.option(
    "--unset",
    is_flag=True,
    help="Instead of exporting all configuration keys, " "this option will unset all keys.",
)
@click.pass_obj
def eval_command(app: ApplicationContext, unset: bool) -> None:
    """Export all configuration values of a configuration for evaluation.

    The output of this command is suitable to be used by a shell's `eval`
    command. It will output the complete configuration as `export key="value"`
    statements, which allow for the preparation of a shell environment.

    Example: eval $(cmemc -c my config eval)

    Warning: Please be aware that credential details are shown in cleartext
    with this command.
    """
    for key in sorted(KNOWN_CONFIG_KEYS):
        if unset:
            app.echo_info(f"unset {key!s}")
        else:
            value = KNOWN_CONFIG_KEYS[key]()
            if value is None or value == "None":
                app.echo_info(f"unset {key!s}")
            else:
                app.echo_info(f'export {key!s}="{value}"')


@click.group(cls=CmemcGroup)
def config() -> CmemcGroup:  # type: ignore[empty-body]
    """List and edit configs as well as get config values.

    Configurations are identified by the section identifier in the
    config file. Each configuration represent a Corporate Memory deployment
    with its specific access method as well as credentials.

    A minimal configuration which uses client credentials has the following
    entries:

    \b
    [example.org]
    CMEM_BASE_URI=https://cmem.example.org/
    OAUTH_GRANT_TYPE=client_credentials
    OAUTH_CLIENT_ID=cmem-service-account
    OAUTH_CLIENT_SECRET=my-secret-account-pass

    Note that OAUTH_GRANT_TYPE can be either client_credentials, password or
    prefetched_token.

    In addition to that, the following config parameters can be used as well:

    \b
    SSL_VERIFY=False    - for ignoring certificate issues (not recommended)
    DP_API_ENDPOINT=URL - to point to a non-standard Explore backend (DataPlatform) location
    DI_API_ENDPOINT=URL - to point to a non-standard Build (DataIntegration) location
    OAUTH_TOKEN_URI=URL - to point to an external IdentityProvider location
    OAUTH_USER=username - only if OAUTH_GRANT_TYPE=password
    OAUTH_PASSWORD=password - only if OAUTH_GRANT_TYPE=password
    OAUTH_ACCESS_TOKEN=token - only if OAUTH_GRANT_TYPE=prefetched_token

    In order to get credential information from an external process, you can
    use the parameter OAUTH_PASSWORD_PROCESS, OAUTH_CLIENT_SECRET_PROCESS and
    OAUTH_ACCESS_TOKEN_PROCESS to set up an external executable.

    \b
    OAUTH_CLIENT_SECRET_PROCESS=/path/to/getpass.sh
    OAUTH_PASSWORD_PROCESS=["getpass.sh", "parameter1", "parameter2"]

    The credential executable can use the cmemc environment for fetching the
    credential (e.g. CMEM_BASE_URI and OAUTH_USER).
    If the credential executable is not given with a full path, cmemc
    will look into your environment PATH for something which can be executed.
    The configured process needs to return the credential on the first line
    of stdout. In addition to that, the process needs to exit with exit
    code 0 (without failure). There are examples available in the online
    manual.
    """  # noqa: D301


config.add_command(list_command)
config.add_command(edit_command)
config.add_command(get_command)
config.add_command(eval_command)



"""Filterable object list."""

import re
from abc import ABC, abstractmethod
from collections.abc import Callable
from inspect import isclass
from typing import Literal

from click import Argument, Context, UsageError
from click.shell_completion import CompletionItem

from cmem_cmemc.completion import finalize_completion, get_completion_args
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.title_helper import TitleHelper


class Filter(ABC):
    """Base class for object list filters"""

    name: str
    description: str

    @abstractmethod
    def is_filtered(self, object_: dict, value: str) -> bool:
        """Return True if the object is filtered (stays in list)."""

    def filter_list(self, objects: list[dict], value: str) -> list[dict]:
        """Filter a given dictionary list"""
        return [o for o in objects if self.is_filtered(object_=o, value=value)]

    @abstractmethod
    def complete_values(self, objects: list[dict], incomplete: str) -> list[CompletionItem]:
        """Provide completion items for filter values"""


def compare_str_equality(ctx: Filter, object_value: str, filter_value: str) -> bool:  # noqa: ARG001
    """Return True if object_value and filter_value are equal"""
    return object_value == filter_value


def compare_int_lower_than(ctx: Filter, object_value: str, filter_value: str) -> bool:
    """Return True if object_value lower than filter_value"""
    try:
        filter_value_int = int(filter_value)
    except ValueError as error:
        raise UsageError(
            f"Invalid filter value '{filter_value}' - need in an integer for filter '{ctx.name}'."
        ) from error
    return int(object_value) < filter_value_int


def compare_int_greater_than(ctx: Filter, object_value: str, filter_value: str) -> bool:
    """Return True if object_value greater than filter_value"""
    try:
        filter_value_int = int(filter_value)
    except ValueError as error:
        raise UsageError(
            f"Invalid filter value '{filter_value}' - need in an integer for filter '{ctx.name}'."
        ) from error
    return int(object_value) > filter_value_int


def compare_regex(ctx: Filter, object_value: str, filter_value: str) -> bool:
    """Return True if object_value matches the regex in filter_value"""
    try:
        pattern = re.compile(filter_value)
    except re.error as error:
        raise UsageError(
            f"Invalid filter value '{filter_value}' - "
            f"need a valid regular expression for filter '{ctx.name}'."
        ) from error
    return bool(re.search(pattern, object_value))


def transform_none(ctx: Filter, value: str) -> str:  # noqa: ARG001
    """Transform: do nothing"""
    return value


def transform_lower(ctx: Filter, value: str) -> str:  # noqa: ARG001
    """Transform: to lower case"""
    return value.lower()


class DirectValuePropertyFilter(Filter):
    """Class to create a filter based on direct properties of an object.

    Missing keys AND Null values are treated as no match (no addition to the completion)
    OR match (completion addition) with the default value if default value is set.
    """

    name: str
    description: str
    property_key: str
    default_value: str | None
    compare: Callable[[Filter, str, str], bool]
    transform: Callable[[Filter, str], str]
    completion_method: Literal["values", "none", "fixed"]
    fixed_completion: list[CompletionItem]
    fixed_completion_only: bool
    title_helper: TitleHelper | None

    def __init__(  # noqa: PLR0913
        self,
        name: str,
        description: str,
        property_key: str,
        default_value: str | None = None,
        compare: Callable[[Filter, str, str], bool] = compare_str_equality,
        transform: Callable[[Filter, str], str] = transform_none,
        completion_method: Literal["values", "none", "fixed"] = "values",
        fixed_completion: list[CompletionItem] | None = None,
        fixed_completion_only: bool = False,
        title_helper: TitleHelper | None = None,
    ):
        """Create the new filter

        name:
            The name of the filter (used as identifier)
        description:
            The description of the filter (used in filter name completion)
        property_key:
            The key of the property which is compared and completed
        default_value:
            Default value which is used in case of missing or NULL values
        compare:
            Overwrite the default string equality comparison function
        transform:
            Overwrite the default value transformation function
        completion_method:
            Overwrite the default completion value provisioning 'values'
        fixed_completion:
            A fixed list of CompletionItem objects used for value completion.
            Initialization with fixed_completion will set completion_method="fixed"
        fixed_completion_only:
            Raise an UsageError if a value is not from fixed completion.
        title_helper:
            (Optional) TitleHelper instance which will be used to provide
            resource titles as descriptions of completions candidates
        """
        self.name = name
        self.description = description
        self.property_key = property_key
        self.default_value = default_value
        self.compare = compare
        self.transform = transform
        self.completion_method = completion_method
        if fixed_completion is not None:
            self.fixed_completion = fixed_completion
            self.completion_method = "fixed"
        else:
            self.fixed_completion = []
        self.fixed_completion_only = fixed_completion_only
        self.title_helper = title_helper

    def is_filtered(self, object_: dict, value: str) -> bool:
        """Return True if the object is filtered (stays in list).

        False in case of missing key, or key is None
        True in case string(key) is value
        """
        filter_value = self.transform(self, value)
        fixed_values = [_.value for _ in self.fixed_completion]
        if self.fixed_completion_only and filter_value not in fixed_values:
            raise UsageError(
                f"'{filter_value}' is not a correct filter value for filter '{self.name}'. "
                f"Use one of {', '.join(fixed_values)}."
            )
        if self.property_key not in object_ or object_[self.property_key] is None:
            if self.default_value is None:
                return False
            return bool(self.compare(self, self.default_value, filter_value))
        object_value = self.transform(self, str(object_[self.property_key]))
        return bool(self.compare(self, object_value, filter_value))

    def complete_values(self, objects: list[dict], incomplete: str) -> list[CompletionItem]:
        """Provide completion items for filter values"""
        if self.completion_method == "none":
            return []
        if self.completion_method == "fixed":
            return self.fixed_completion
        if self.completion_method == "values":
            candidates: list = []
            for _ in objects:
                if self.property_key not in _ or _[self.property_key] is None:
                    if self.default_value is None:
                        # without actual values and no defaults, we can not complete something
                        continue
                    # without actual values but at least a default value
                    candidate = self.transform(self, str(self.default_value))
                else:
                    # a normal candidate value
                    candidate = self.transform(self, str(_[self.property_key]))
                candidates.append(candidate)
            if self.title_helper:
                self.title_helper.get(list(set(candidates)))
                candidates = [(str(_), self.title_helper.get(_)) for _ in candidates]
            return finalize_completion(
                candidates=candidates,
                incomplete=incomplete,
            )
        raise NotImplementedError(f"Completion method {self.completion_method} not implemented.")


class DirectListPropertyFilter(Filter):
    """Class to create filter based on direct list properties of an object

    False in case of missing key, key is None, or value is not in list
    True in case value is in the list
    """

    name: str
    description: str
    property_key: str
    title_helper: TitleHelper | None

    def __init__(
        self,
        name: str,
        description: str,
        property_key: str,
        title_helper: TitleHelper | None = None,
    ):
        """Create the new filter

        name:
            The name of the filter (used as identifier)
        description:
            The description of the filter (used in filter name completion)
        property_key:
            The key of the property which is compared and completed
        title_helper:
            (Optional) TitleHelper instance which will be used to provide
            resource titles as descriptions of completions candidates
        """
        self.name = name
        self.description = description
        self.property_key = property_key
        self.title_helper = title_helper

    def is_filtered(self, object_: dict, value: str) -> bool:
        """Return True if the object is filtered (stays in list)."""
        if self.property_key not in object_:
            return False  # key is not in object
        if object_[self.property_key] is None:
            return False  # key value is None
        if not isinstance(object_[self.property_key], list):
            return False  # key value is not a list
        return value in [str(_) for _ in object_[self.property_key]]

    def complete_values(self, objects: list[dict], incomplete: str) -> list[CompletionItem]:
        """Provide completion items for filter values"""
        candidates: list = []
        for object_ in objects:
            if self.property_key not in object_:
                continue  # key is not in object
            if object_[self.property_key] is None:
                continue  # key value is None
            if not isinstance(object_[self.property_key], list):
                continue  # key value is not a list
            candidates.extend([str(_) for _ in object_[self.property_key]])
        if self.title_helper:
            self.title_helper.get(list(set(candidates)))
            candidates = [(_, self.title_helper.get(_)) for _ in candidates]
        return finalize_completion(candidates=candidates, incomplete=incomplete)


class ObjectList:
    """Filterable object list"""

    name: str
    filters: dict[str, Filter]
    get_objects: Callable[[Context], list[dict]]

    def __init__(
        self,
        get_objects: Callable[[Context], list[dict]],
        filters: list[Filter | type[Filter]] | None = None,
        name: str = "list",
    ) -> None:
        self.get_objects = get_objects
        self.filters = {}
        self.name = name
        if filters:
            for _ in filters:
                self.add_filter(_)

    def add_filter(self, filter_: Filter | type[Filter]) -> None:
        """Add a filter to the object list"""
        added_filter = filter_() if isclass(filter_) else filter_
        if not isinstance(added_filter, Filter):
            raise TypeError("'filter_' parameter must be an instance OR a subclass of Filter")
        if added_filter.name in self.filters:
            raise UsageError(f"Filter {added_filter.name} already exists")
        self.filters[added_filter.name] = added_filter

    def remove_filter(self, filter_name: str) -> None:
        """Remove a filter from the object list"""
        self.filters.pop(filter_name)

    def purge_filters(self) -> None:
        """Remove all filters from the object list"""
        self.filters = {}

    def get_filter_help_text(self) -> str:
        """Get help text for the filter option."""
        return (
            f"Filter {self.name} by one of the following filter names and a "
            f"corresponding value: {', '.join(self.filters.keys())}."
        )

    def get_filter_names(self) -> list[str]:
        """Get names of all filters added to the object list"""
        return list(self.filters.keys())

    def get_filter(self, name: str) -> Filter:
        """Get filter by name"""
        try:
            return self.filters[name]
        except KeyError as error:
            filter_name_list = ", ".join(self.get_filter_names())
            raise UsageError(f"Invalid filter name - use one of {filter_name_list}") from error

    def apply_filters(
        self,
        ctx: Context,
        filter_: tuple[tuple[str, str]] | list[tuple[str, str]] | None = None,
        objects: list[dict] | None = None,
    ) -> list[dict]:
        """Filter a given object list"""
        filtered = list(self.get_objects(ctx)) if objects is None else list(objects)
        if not filter_:
            return filtered
        for filter_name, filter_value in filter_:
            the_filter = self.get_filter(filter_name)
            filtered = the_filter.filter_list(value=filter_value, objects=filtered)
        return filtered

    def complete_values(
        self,
        ctx: Context,
        param: Argument,  # noqa: ARG002
        incomplete: str,
    ) -> list[CompletionItem]:
        """Complete names and values for the filter collection"""
        previous_filter = ctx.params.get("filter_")  # tuple of (name, value) pairs or NONE
        previous_filter = previous_filter if previous_filter is not None else []
        previous_filter_names = [_[0] for _ in previous_filter]
        args = get_completion_args(incomplete)
        last_argument = args[len(args) - 1]

        if last_argument == "--filter":
            # complete filter names and descriptions
            candidates = [
                (name, self.get_filter(name).description)
                for name in self.get_filter_names()
                if name not in previous_filter_names  # do not show already used filters
            ]
            return finalize_completion(
                candidates=candidates,
                incomplete=incomplete,
            )

        ApplicationContext.set_connection_from_params(ctx.find_root().params)
        # This will filter the object list with name/values filter from the command line,
        # up to the current filter which values are completed
        objects = self.apply_filters(ctx=ctx, filter_=previous_filter)
        # provide completion for the values of the current filter
        return self.get_filter(last_argument).complete_values(
            objects=objects, incomplete=incomplete
        )



"""Build scheduler commands for the cmem command line interface."""

from typing import Any

import click
from click import Argument, ClickException, UsageError
from cmem.cmempy.config import get_cmem_base_uri
from cmem.cmempy.workflow.workflow import get_workflow_editor_uri
from cmem.cmempy.workspace.search import list_items
from cmem.cmempy.workspace.tasks import get_task, patch_parameter

from cmem_cmemc import completion
from cmem_cmemc.command import CmemcCommand
from cmem_cmemc.command_group import CmemcGroup
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.utils import split_task_id, struct_to_table


def tuple_to_list(ctx: ApplicationContext, param: Argument, value: tuple) -> list:  # noqa: ARG001
    """Get a list from a tuple

    Used as callback to have mutable values
    """
    return list(value)


def _get_schedulers() -> list[dict[Any, Any]]:
    """Return the exported list of schedulers."""
    return_value: list[dict[Any, Any]] = list_items(
        item_type="task",
        add_task_parameter=True,
        facets=[{"facetId": "taskType", "keywordIds": ["Scheduler"], "type": "keyword"}],
    )["results"]
    return return_value


def _get_sorted_scheduler_ids() -> list[str]:
    """Return sorted list of scheduler IDs."""
    return sorted(
        [_["projectId"] + ":" + _["id"] for _ in _get_schedulers()], key=lambda k: k[0].lower()
    )


@click.command(cls=CmemcCommand, name="open")
@click.argument(
    "scheduler_ids",
    nargs=-1,
    required=True,
    type=click.STRING,
    shell_complete=completion.scheduler_ids,
)
@click.option(
    "--workflow",
    is_flag=True,
    help="Instead of opening the scheduler page, open the page of the " "scheduled workflow.",
)
@click.pass_obj
def open_command(app: ApplicationContext, scheduler_ids: tuple[str, ...], workflow: bool) -> None:
    """Open scheduler(s) in the browser.

    With this command, you can open a scheduler in the workspace in
    your browser to change it.

    The command accepts multiple scheduler IDs which results in
    opening multiple browser tabs.
    """
    schedulers = _get_schedulers()
    all_scheduler_ids = [s["projectId"] + ":" + s["id"] for s in schedulers]
    for scheduler_id in scheduler_ids:
        if scheduler_id not in all_scheduler_ids:
            raise ClickException(f"Scheduler '{scheduler_id}' not found.")
    for scheduler_id in scheduler_ids:
        for _ in schedulers:
            current_id = _["projectId"] + ":" + _["id"]
            if scheduler_id == current_id:
                if workflow:
                    uri = get_workflow_editor_uri().format(_["projectId"], _["parameters"]["task"])
                else:
                    uri = get_cmem_base_uri() + _["itemLinks"][0]["path"]
                app.echo_debug(f"Open {_}: {uri}")
                click.launch(uri)


@click.command(cls=CmemcCommand, name="list")
@click.option("--raw", is_flag=True, help="Outputs raw JSON.")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only task identifier and no labels or other "
    "metadata. This is useful for piping the IDs into other commands.",
)
@click.pass_obj
def list_command(app: ApplicationContext, raw: bool, id_only: bool) -> None:
    """List available scheduler.

    Outputs a table or a list of scheduler IDs which can be used as
    reference for the scheduler commands.
    """
    schedulers = _get_schedulers()
    if raw:
        app.echo_info_json(schedulers)
        return
    if id_only:
        for _ in sorted(s["projectId"] + ":" + s["id"] for s in schedulers):
            app.echo_result(_)
        return
    # output a user table
    table = []
    headers = ["Scheduler ID", "Interval", "Enabled", "Label"]
    for _ in schedulers:
        row = [
            _["projectId"] + ":" + _["id"],
            _["parameters"]["interval"],
            _["parameters"]["enabled"],
            _["label"],
        ]
        table.append(row)
    # sort output by label - https://docs.python.org/3/howto/sorting.html
    app.echo_info_table(
        table,
        headers=headers,
        sort_column=1,
        empty_table_message="Now workflow scheduler found. "
        "Open a project in the web interface and create a new workflow scheduler there.",
    )


@click.command(cls=CmemcCommand, name="inspect")
@click.argument("scheduler_id", type=click.STRING, shell_complete=completion.scheduler_ids)
@click.option("--raw", is_flag=True, help="Outputs raw JSON.")
@click.pass_obj
def inspect_command(app: ApplicationContext, scheduler_id: str, raw: bool) -> None:
    """Display all metadata of a scheduler."""
    app.echo_debug(f"Scheduler ID is {scheduler_id}")
    project_id, scheduler_id = split_task_id(scheduler_id)
    task = get_task(project_id, scheduler_id)
    if raw:
        app.echo_info_json(task)
    else:
        table = struct_to_table(task)
        app.echo_info_table(table, headers=["Key", "Value"], sort_column=0)


@click.command(cls=CmemcCommand, name="disable")
@click.argument(
    "scheduler_ids",
    nargs=-1,
    type=click.STRING,
    shell_complete=completion.scheduler_ids,
    callback=tuple_to_list,
)
@click.option("-a", "--all", "all_", is_flag=True, help="Disable all scheduler.")
@click.pass_obj
def disable_command(app: ApplicationContext, scheduler_ids: list[str], all_: bool) -> None:
    """Disable scheduler(s).

    The command accepts multiple scheduler IDs which results in disabling
    them one after the other.
    """
    if not scheduler_ids and not all_:
        raise UsageError(
            "Either specify at least one scheduler ID or use "
            "the --all option to disable all scheduler."
        )
    if all_:
        scheduler_ids = _get_sorted_scheduler_ids()

    app.echo_debug(f"Scheduler IDs are {scheduler_ids}")
    count = len(scheduler_ids)
    for current, _ in enumerate(scheduler_ids, start=1):
        current_string = str(current).zfill(len(str(count)))
        project_id, scheduler_id = split_task_id(_)
        task = get_task(project_id, scheduler_id, with_labels=False)
        app.echo_info(
            f"Disable scheduler {current_string}/{count}: " f"{project_id}:{scheduler_id} ... ",
            nl=False,
        )
        data = {"data": {"parameters": task["data"]["parameters"]}}
        if data["data"]["parameters"]["enabled"] == "false":
            app.echo_warning("already disabled")
            continue
        data["data"]["parameters"]["enabled"] = "false"
        patch_parameter(project=project_id, task=scheduler_id, data=data)
        app.echo_success("done")


@click.command(cls=CmemcCommand, name="enable")
@click.argument(
    "scheduler_ids",
    nargs=-1,
    type=click.STRING,
    shell_complete=completion.scheduler_ids,
    callback=tuple_to_list,
)
@click.option("-a", "--all", "all_", is_flag=True, help="Enable all scheduler.")
@click.pass_obj
def enable_command(app: ApplicationContext, scheduler_ids: list[str], all_: bool) -> None:
    """Enable scheduler(s).

    The command accepts multiple scheduler IDs which results in enabling
    them one after the other.
    """
    if not scheduler_ids and not all_:
        raise UsageError(
            "Either specify at least one scheduler ID or use "
            "the --all option to enable all scheduler."
        )
    if all_:
        scheduler_ids = _get_sorted_scheduler_ids()

    app.echo_debug(f"Scheduler IDs are {scheduler_ids}")
    count = len(scheduler_ids)
    for current, _ in enumerate(scheduler_ids):
        current_string = str(current).zfill(len(str(count)))
        project_id, scheduler_id = split_task_id(_)
        task = get_task(project_id, scheduler_id, with_labels=False)
        app.echo_info(
            f"Enable scheduler {current_string}/{count}: " f"{project_id}:{scheduler_id} ... ",
            nl=False,
        )
        data = {"data": {"parameters": task["data"]["parameters"]}}
        if data["data"]["parameters"]["enabled"] == "true":
            app.echo_warning("already enabled")
            continue
        data["data"]["parameters"]["enabled"] = "true"
        patch_parameter(project=project_id, task=scheduler_id, data=data)
        app.echo_success("done")


@click.group(cls=CmemcGroup)
def scheduler() -> CmemcGroup:  # type: ignore[empty-body]
    """List, inspect, enable/disable or open scheduler.

    Schedulers execute workflows in specified intervals. They are identified
    with a SCHEDULER_ID. To get a list of existing schedulers, execute the
    list command or use tab-completion.
    """


scheduler.add_command(open_command)
scheduler.add_command(list_command)
scheduler.add_command(inspect_command)
scheduler.add_command(disable_command)
scheduler.add_command(enable_command)



"""Title helper functions."""

import json

from cmem.cmempy.api import get_json
from cmem.cmempy.config import get_dp_api_endpoint


class TitleHelper:
    """Title helper class."""

    fetched_labels: dict[str, dict]

    def __init__(self) -> None:
        self.fetched_labels = {}
        self.endpoint = f"{get_dp_api_endpoint()}/api/explore/titles"

    def get(self, iri: str | list[str]) -> str | dict[str, str]:
        """Get the title of an IRI (or list of IRI)."""
        output = {}
        iris = [iri] if isinstance(iri, str) else list(set(iri))

        iris_to_fetch = []
        for _ in iris:
            if _ in self.fetched_labels:
                output[_] = self.fetched_labels[_]["title"]
            else:
                iris_to_fetch.append(_)

        if len(iris_to_fetch) > 0:
            titles: dict = get_json(
                self.endpoint,
                method="POST",
                data=json.dumps(iris_to_fetch),
                headers={"Content-type": "application/json"},
            )
            for title in titles.values():
                self.fetched_labels[title["iri"]] = title
                output[title["iri"]] = title["title"]

        return output[iri] if isinstance(iri, str) else output



"""Keycloak user management commands"""

import sys
from getpass import getpass

import click
from click import ClickException
from cmem.cmempy.config import get_keycloak_base_uri, get_keycloak_realm_id
from cmem.cmempy.keycloak.group import list_groups
from cmem.cmempy.keycloak.user import (
    assign_groups,
    create_user,
    delete_user,
    get_user_by_username,
    list_users,
    request_password_change,
    reset_password,
    unassign_groups,
    update_user,
    user_groups,
)

from cmem_cmemc import completion
from cmem_cmemc.command import CmemcCommand
from cmem_cmemc.command_group import CmemcGroup
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.object_list import (
    DirectValuePropertyFilter,
    ObjectList,
    compare_regex,
    transform_lower,
)

NO_USER_ERROR = (
    "{} is not a valid user account. Use the 'admin user list' command "
    "to get a list of existing user accounts."
)
EXISTING_USER_ERROR = "{} does already exist"
NO_GROUP_ERROR = "{} is not a valid group. Valid groups are {}"
INVALID_UNASSIGN_GROUP_ERROR = "Group {} is not assigned to user. Valid groups are {}"
NO_EMAIL_ERROR = "Email is empty for {} user."


def get_users(ctx: click.Context) -> list[dict]:  # noqa: ARG001
    """Get users for object list"""
    users: list[dict] = list_users()
    return users


user_list = ObjectList(
    name="users",
    get_objects=get_users,
    filters=[
        DirectValuePropertyFilter(
            name="enabled",
            description="Filter accounts by enabled flag.",
            property_key="enabled",
            transform=transform_lower,
        ),
        DirectValuePropertyFilter(
            name="email",
            description="Filter accounts by regex matching the email address.",
            property_key="email",
            compare=compare_regex,
            fixed_completion=[],
        ),
        DirectValuePropertyFilter(
            name="username",
            description="Filter accounts by regex matching the username.",
            property_key="username",
            compare=compare_regex,
            fixed_completion=[],
        ),
    ],
)


@click.command(cls=CmemcCommand, name="list")
@click.option("--raw", is_flag=True, help="Outputs raw JSON.")
@click.option(
    "--filter",
    "filter_",
    type=(str, str),
    multiple=True,
    help=user_list.get_filter_help_text(),
    shell_complete=user_list.complete_values,
)
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only username. " "This is useful for piping the IDs into other commands.",
)
@click.pass_context
def list_command(
    ctx: click.Context, filter_: tuple[tuple[str, str]], raw: bool, id_only: bool
) -> None:
    """List user accounts.

    Outputs a list of user accounts, which can be used to get an overview as well
    as a reference for the other commands of the `admin user` command group.
    """
    app = ctx.obj
    users = user_list.apply_filters(ctx=ctx, filter_=filter_)
    if raw:
        app.echo_info_json(users)
        return
    if id_only:
        for usr in users:
            app.echo_info(usr["username"])
        return
    table = [
        (
            usr["username"],
            usr.get("firstName", "-"),
            usr.get("lastName", "-"),
            usr.get("email", "-"),
        )
        for usr in users
    ]
    app.echo_info_table(
        table,
        headers=["Username", "First Name", "Last Name", "Email"],
        sort_column=0,
        empty_table_message="No user accounts found. "
        "Use the `admin user create` command to create an account.",
    )


@click.command(cls=CmemcCommand, name="delete")
@click.argument("username", shell_complete=completion.user_ids)
@click.pass_obj
def delete_command(app: ApplicationContext, username: str) -> None:
    """Delete a user account.

    This command deletes a user account from a realm.

    Note: The deletion of a user account does not delete the assigned groups of
    this account, only the assignments to these groups.
    """
    app.echo_info(f"Deleting user {username} ... ", nl=False)
    users = get_user_by_username(username)
    if not users:
        raise ClickException(NO_USER_ERROR.format(username))

    delete_user(users[0]["id"])
    app.echo_success("deleted")


@click.command(cls=CmemcCommand, name="create")
@click.argument("username")
@click.pass_obj
def create_command(app: ApplicationContext, username: str) -> None:
    """Create a user account.

    This command creates a new user account.

    Note: The created user account has no metadata such as personal data or group
    assignments. In order to add these details to a user account, use the
    `admin user update` command.
    """
    app.echo_info(f"Creating user {username} ... ", nl=False)
    users = get_user_by_username(username)
    if users:
        raise ClickException(EXISTING_USER_ERROR.format(username))

    create_user(username=username)
    app.echo_success("done")


@click.command(cls=CmemcCommand, name="update")
@click.argument("username", shell_complete=completion.user_ids)
@click.option("--first-name", type=click.STRING, required=False, help="Set a new first name.")
@click.option("--last-name", type=click.STRING, required=False, help="Set a new last name.")
@click.option("--email", type=click.STRING, required=False, help="Set a new email.")
@click.option(
    "--assign-group",
    type=click.STRING,
    multiple=True,
    shell_complete=completion.user_group_ids,
    help="Assign a group.",
)
@click.option(
    "--unassign-group",
    type=click.STRING,
    multiple=True,
    shell_complete=completion.user_group_ids,
    help="Unassign a group.",
)
@click.pass_obj
def update_command(  # noqa: PLR0913
    app: ApplicationContext,
    username: str,
    first_name: str,
    last_name: str,
    email: str,
    assign_group: tuple[str, ...],
    unassign_group: tuple[str, ...],
) -> None:
    """Update a user account.

    This command updates metadata and group assignments of a user account.

    For each data value, a separate option needs to be used. All options can
    be combined in a single execution.

    Note: In order to assign a group to a user account, the group need to be
    added or imported to the realm upfront.
    """
    options = (first_name, last_name, email, assign_group, unassign_group)
    if all(_ is None or _ == () for _ in options):
        raise click.UsageError(
            "This commands needs to be used with at least one option "
            "(e.g. --email). See the command help for a list of options."
        )
    app.echo_info(f"Updating user {username} ... ", nl=False)
    users = get_user_by_username(username)
    if not users:
        raise ClickException(NO_USER_ERROR.format(username))
    user_id = users[0]["id"]
    all_groups = {group["name"]: group["id"] for group in list_groups()}
    invalid_groups = [group for group in assign_group if group not in all_groups]
    existing_user_groups = {group["name"] for group in user_groups(user_id)}

    if invalid_groups:
        raise ClickException(
            NO_GROUP_ERROR.format(
                ", ".join(invalid_groups), ", ".join(all_groups.keys() - set(existing_user_groups))
            )
        )

    invalid_unassign_groups = [
        group for group in unassign_group if group not in existing_user_groups
    ]
    if invalid_unassign_groups:
        raise ClickException(
            INVALID_UNASSIGN_GROUP_ERROR.format(
                ", ".join(invalid_unassign_groups), ", ".join(existing_user_groups)
            )
        )

    assign_group_ids = [all_groups[name] for name in assign_group]
    unassign_groups_ids = [all_groups[name] for name in unassign_group]
    unassign_groups(user_id, unassign_groups_ids)
    assign_groups(user_id, assign_group_ids)

    update_user(
        user_id=user_id,
        username=username,
        first_name=first_name,
        last_name=last_name,
        email=email,
        email_verified=True if email else None,
    )
    app.echo_success("done")


@click.command(cls=CmemcCommand, name="password")
@click.argument("username", shell_complete=completion.user_ids)
@click.option(
    "--value", help="With this option, the new password can be set in a non-interactive way."
)
@click.option(
    "--temporary", is_flag=True, help="If enabled, the user must change the password on next login."
)
@click.option(
    "--request-change",
    is_flag=True,
    help="If enabled, will send a email to user to reset the password.",
)
@click.pass_obj
def password_command(
    app: ApplicationContext, username: str, value: str, temporary: bool, request_change: bool
) -> None:
    """Change the password of a user account.

    With this command, the password of a user account can be changed.
    The default execution mode of this command is an interactive prompt which asks
    for the password (twice). In order automate password changes, you can use the
    `--value` option.

    Warning: Providing passwords on the command line can be dangerous
    (e.g. due to a potential exploitation in the shell history).
    A suggested more save way for automation is to provide the password in a variable
    first (e.g. with `NEW_PASS=$(pwgen -1 40)`) and use it afterward in the
    cmemc call: `cmemc admin user password max --value ${NEW_PASS}`.
    """
    app.echo_info(f"Changing password for account {username} ... ", nl=False)
    users = get_user_by_username(username)
    if not users:
        raise ClickException(NO_USER_ERROR.format(username))
    if not value and not request_change:
        app.echo_info("\nNew password: ", nl=False)
        value = getpass(prompt="")
        app.echo_info("Retype new password: ", nl=False)
        retype_password = getpass(prompt="")
        if value != retype_password:
            app.echo_error("Sorry, passwords do not match.")
            app.echo_error("password unchanged")
            sys.exit(1)
    if value:
        reset_password(user_id=users[0]["id"], value=value, temporary=temporary)
    if request_change and not users[0].get("email", None):
        raise ClickException(NO_EMAIL_ERROR.format(username))
    if request_change:
        request_password_change(users[0]["id"])
    app.echo_success("done")


@click.command(cls=CmemcCommand, name="open")
@click.argument(
    "usernames", nargs=-1, required=False, type=click.STRING, shell_complete=completion.user_ids
)
@click.pass_obj
def open_command(app: ApplicationContext, usernames: str) -> None:
    """Open user in the browser.

    With this command, you can open a user in the keycloak console in
    your browser to change them.

    The command accepts multiple usernames which results in
    opening multiple browser tabs.
    """
    open_user_base_uri = (
        f"{get_keycloak_base_uri()}/admin/master/console/#/" f"{get_keycloak_realm_id()}/users"
    )
    if not usernames:
        app.echo_debug(f"Open users list: {open_user_base_uri}")
        click.launch(open_user_base_uri)
    else:
        users = list_users()
        user_name_id_map = {u["username"]: u["id"] for u in users}
        for _ in usernames:
            if _ not in user_name_id_map:
                raise ClickException(NO_USER_ERROR.format(_))
            user_id = user_name_id_map[_]
            open_user_uri = f"{open_user_base_uri}/{user_id}/settings"

            app.echo_debug(f"Open {_}: {open_user_uri}")
            click.launch(open_user_uri)


@click.group(cls=CmemcGroup)
def user() -> CmemcGroup:  # type: ignore[empty-body]
    """List, create, delete and modify user accounts.

    This command group is an opinionated interface to the Keycloak realm of your
    Corporate Memory instance. In order to be able to manage user data, the
    configured cmemc connection account needs to be equipped with the
    `manage-users` role in the used realm.

    User accounts are identified by a username which unique in the scope of
    the used realm.

    In case your Corporate Memory deployment does not use the default deployment
    layout, the following additional config variables can be used in your
    connection configuration: `KEYCLOAK_BASE_URI` defaults to
    `/auth` on `CMEM_BASE_URI` and locates your Keycloak deployment;
    `KEYCLOAK_REALM_ID` defaults to `cmem` and identifies the used realm.
    """


user.add_command(list_command)
user.add_command(create_command)
user.add_command(update_command)
user.add_command(delete_command)
user.add_command(password_command)
user.add_command(open_command)



"""workspace commands for cmem command line interface."""

import os

import click
from cmem.cmempy.workspace import reload_workspace
from cmem.cmempy.workspace.export_ import export
from cmem.cmempy.workspace.import_ import import_workspace
from jinja2 import Template

from cmem_cmemc import completion
from cmem_cmemc.command import CmemcCommand
from cmem_cmemc.command_group import CmemcGroup
from cmem_cmemc.commands.python import python
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.parameter_types.path import ClickSmartPath
from cmem_cmemc.smart_path import SmartPath as Path


@click.command(cls=CmemcCommand, name="export")
@click.option(
    "-o",
    "--overwrite",
    is_flag=True,
    hidden=True,
)
@click.option(
    "--replace",
    is_flag=True,
    help="Replace existing files. This is a dangerous option, so use it with care.",
)
@click.option(
    "--type",
    "marshalling_plugin",
    default="xmlZip",
    show_default=True,
    type=click.STRING,
    shell_complete=completion.marshalling_plugins,
    help="Type of the exported workspace file.",
)
@click.option(
    "--filename-template",
    "-t",
    "template",
    default="{{date}}-{{connection}}.workspace",
    show_default=True,
    type=click.STRING,
    shell_complete=completion.workspace_export_templates,
    help="Template for the export file name. "
    "Possible placeholders are (Jinja2): "
    "{{connection}} (from the --connection option) and "
    "{{date}} (the current date as YYYY-MM-DD). "
    "The file suffix will be appended. "
    "Needed directories will be created.",
)
@click.argument(
    "file",
    shell_complete=completion.workspace_files,
    required=False,
    type=ClickSmartPath(writable=True, allow_dash=False, dir_okay=False),
)
@click.pass_obj
def export_command(  # noqa: PLR0913
    app: ApplicationContext,
    overwrite: bool,
    replace: bool,
    marshalling_plugin: str,
    template: str,
    file: str,
) -> None:
    """Export the complete workspace (all projects) to a ZIP file.

    Depending on the requested export type, this ZIP file contains either one
    Turtle file per project (type `rdfTurtle`) or a substructure of resource
    files and XML descriptions (type `xmlZip`).

    The file name is optional and will be generated with by
    the template if absent.
    """
    if overwrite:
        replace = overwrite
        app.echo_warning(
            "The option --overwrite is deprecated and will be removed with the next major release."
            " Please use the --replace option instead."
        )
    if file is None:
        # prepare the template data and create the actual file incl. suffix
        template_data = app.get_template_data()
        file = Template(template).render(template_data) + ".zip"
    file = os.path.normpath(file)
    app.echo_info(f"Export workspace to {file} ... ", nl=False)
    if Path(file).exists() and replace is not True:
        app.echo_error("file exists")
    else:
        # output directory is created lazy
        Path(file).parent.mkdir(parents=True, exist_ok=True)
        # do the export
        export_data = export(marshalling_plugin)
        with Path(file).open(mode="wb") as export_file:
            export_file.write(export_data)
        app.echo_success("done")


@click.command(cls=CmemcCommand, name="import")
@click.option(
    "--type",
    "marshalling_plugin",
    default="xmlZip",
    show_default=True,
    type=click.STRING,
    shell_complete=completion.marshalling_plugins,
    help="Type of the exported workspace file.",
)
@click.argument(
    "file",
    shell_complete=completion.workspace_files,
    type=ClickSmartPath(readable=True, allow_dash=False, dir_okay=False),
)
@click.pass_obj
def import_command(app: ApplicationContext, file: str, marshalling_plugin: str) -> None:
    """Import the workspace from a file."""
    app.echo_info(f"Import workspace from {file} ... ", nl=False)
    import_workspace(file, marshalling_plugin)
    app.echo_success("done")


@click.command(cls=CmemcCommand, name="reload")
@click.pass_obj
def reload_command(app: ApplicationContext) -> None:
    """Reload the workspace from the backend."""
    app.echo_info("Reload workspace  ... ", nl=False)
    reload_workspace()
    app.echo_success("done")


@click.group(cls=CmemcGroup)
def workspace() -> CmemcGroup:  # type: ignore[empty-body]
    """Import, export and reload the project workspace."""


workspace.add_command(export_command)
workspace.add_command(import_command)
workspace.add_command(reload_command)
workspace.add_command(python)



"""Provides client classes for interacting with different storage systems."""

from __future__ import annotations

import os
import urllib.parse
from pathlib import Path
from typing import IO, TYPE_CHECKING, ClassVar

import smart_open

from cmem_cmemc.smart_path.clients.http import HttpPath

if TYPE_CHECKING:
    from collections.abc import Generator

    from cmem_cmemc.smart_path.clients import StoragePath


class SmartPath:
    """Smart path"""

    SUPPORTED_SCHEMAS: ClassVar = {
        "file": Path,
        "http": HttpPath,
        "https": HttpPath,
    }

    def __init__(self, path: str):
        self.path = path
        self.schema = self._sniff_schema(self.path)
        if self.schema not in self.SUPPORTED_SCHEMAS:
            raise NotImplementedError(f"Schema '{self.schema}' not supported")
        self._client: StoragePath = self.SUPPORTED_SCHEMAS.get(self.schema)(self.path)

    @staticmethod
    def _sniff_schema(path: str) -> str:
        """Return the scheme of the URL only, as a string."""
        #
        # urlsplit doesn't work on Windows -- it parses the drive as the scheme...
        # no protocol given => assume a local file
        #
        if os.name == "nt" and "://" not in path:
            path = "file://" + path
        schema = urllib.parse.urlsplit(path).scheme
        return schema if schema else "file"

    def is_dir(self) -> bool:
        """Determine if path is a directory or not."""
        return self._client.is_dir()

    def is_file(self) -> bool:
        """Return the suffix of the path."""
        return self._client.is_file()

    def exists(self) -> bool:
        """Determine if path exists or not."""
        return self._client.exists()

    @property
    def suffix(self) -> str:
        """Return the suffix of the path."""
        return self._client.suffix

    @property
    def parent(self) -> StoragePath:
        """The logical parent of the path."""
        return self._client.parent

    @property
    def name(self) -> str:
        """Determine the name of the path."""
        return self._client.name

    def open(self, mode: str = "r", encoding: str | None = None) -> IO:
        """Open the file pointed by this path."""
        file: IO = smart_open.open(self.path, mode=mode, encoding=encoding)
        return file

    def mkdir(self, parents: bool = False, exist_ok: bool = False) -> None:
        """Return the suffix of the path."""
        self._client.mkdir(parents=parents, exist_ok=exist_ok)

    def glob(self, pattern: str) -> Generator[StoragePath, StoragePath, StoragePath]:
        """Iterate over this subtree and yield all existing files"""
        return self._client.glob(pattern=pattern)

    def resolve(self) -> StoragePath:
        """Iterate over this subtree and yield all existing files"""
        return self._client.resolve()

    def __truediv__(self, key: str) -> StoragePath:
        """Return StoragePath with appending the key to the exising path"""
        return self._client.__truediv__(key)



"""Utility functions for CLI interface."""

import json
import os
import pathlib
import re
import sys
import unicodedata
from dataclasses import dataclass
from importlib.metadata import version as cmemc_version
from typing import TYPE_CHECKING
from zipfile import BadZipFile, ZipFile

import requests
from click import Argument, ClickException
from cmem.cmempy.dp.proxy.graph import get_graphs_list
from cmem.cmempy.queries import QueryCatalog
from cmem.cmempy.workspace.projects.project import get_projects
from prometheus_client import Metric

from cmem_cmemc.config_parser import PureSectionConfigParser
from cmem_cmemc.constants import NAMESPACES
from cmem_cmemc.smart_path import SmartPath

if TYPE_CHECKING:
    from cmem_cmemc.context import ApplicationContext


def get_version() -> str:
    """Get the current version or SNAPSHOT."""
    return cmemc_version("cmem-cmemc")


def check_python_version(ctx: type["ApplicationContext"]) -> None:
    """Check the runtime python version and warn or error."""
    version = sys.version_info
    major_expected = [3]
    minor_expected = [10, 11, 12, 13]
    if version.major not in major_expected:
        ctx.echo_error(f"Error: cmemc can not be executed with Python {version.major}.")
        sys.exit(1)
    if version.minor not in minor_expected and not ctx.is_completing():
        ctx.echo_warning(
            "Warning: You are running cmemc under a non-tested python "
            f"environment ({version.major}.{version.minor})."
        )


def extract_error_message(error: Exception, ignore_error_prefix: bool = False) -> str:
    """Extract a message from an exception."""
    # exceptions with response is HTTPError
    message = type(error).__name__ + ": " + str(error) + "\n"
    if ignore_error_prefix:
        message = str(error) + "\n"

    try:
        # try to load Problem Details for HTTP API JSON
        details = json.loads(error.response.text)  # type: ignore[attr-defined]
        message += type(error).__name__ + ": "
        if "title" in details:
            message += details["title"] + ": "
        if "detail" in details:
            message += details["detail"]
    except (AttributeError, ValueError):
        # is not json or any other issue, output plain response text
        pass
    return message.strip()


def iri_to_qname(iri: str) -> str:
    """Return a qname for an IRI based on well known namespaces.

    In case of no matching namespace, the full IRI is returned.

    Args:
    ----
        iri: the IRI to transform

    Returns: string

    """
    for prefix, namespace in NAMESPACES.items():
        iri = iri.replace(namespace, prefix + ":")
    return iri


@dataclass
class RdfGraphData:
    """Represents the data structure for RDF graph information.

    Attributes:
        file_path: The absolute path to the RDF data file.
        graph_iri: The iri of the graph.
        graph_imports: A list of graph imports.

    """

    file_path: str
    graph_iri: str
    graph_imports: list[str]


def read_rdf_graph_files(directory_path: str) -> list[RdfGraphData]:
    """Read all files from directory_path and output as RdfGraphData."""
    rdf_graphs: list[RdfGraphData] = []
    for root, _, files in os.walk(directory_path):
        for _file in files:
            if _file.endswith((".graph", ".imports")):
                continue
            file_path = SmartPath(root) / _file
            # Handle compressed files (like .gz)
            if _file.endswith(".gz"):
                _graph_file = _file.replace(".gz", ".graph")
                _graph_imports_file = _file.replace(".gz", ".imports")
            else:
                _graph_file = f"{_file}.graph"
                _graph_imports_file = f"{_file}.imports"
            graph_file_path = SmartPath(root) / _graph_file
            imports_file_path = SmartPath(root) / _graph_imports_file
            graph_name = ""
            graph_imports = []
            if graph_file_path.exists():
                graph_name = read_file_to_string(str(graph_file_path)).strip()

            if imports_file_path.exists():
                # Read the graph imports.
                imports_content = read_file_to_string(str(imports_file_path)).strip()
                graph_imports = imports_content.split("\n") if imports_content else []

            if graph_name:
                rdf_graphs.append(
                    RdfGraphData(
                        file_path=str(file_path.resolve()),
                        graph_iri=graph_name,
                        graph_imports=graph_imports,
                    )
                )
    return rdf_graphs


def read_file_to_string(file_path: str) -> str:
    """Read file to string."""
    with SmartPath(file_path).open(mode="rb") as _file:
        return str(_file.read().decode("utf-8"))


def get_graphs(writeable: bool = True, readonly: bool = True) -> list:
    """Retrieve list of accessible graphs from DP endpoint.

    readonly=True|writeable=True outputs all graphs
    readonly=False|writeable=True outputs only writeable graphs
    readonly=True|writeable=False outputs graphs without write access
    (but read access)
    """
    all_graphs = get_graphs_list()
    filtered_graphs = []
    for graph in all_graphs:
        if graph["writeable"] and writeable:
            filtered_graphs.append(graph)
        if not graph["writeable"] and readonly:
            filtered_graphs.append(graph)
    return filtered_graphs


def get_graphs_as_dict(writeable: bool = True, readonly: bool = True) -> dict:
    """Get the graph response as dict with IRI as main key."""
    graph_dict = {}
    for graph in get_graphs(writeable=writeable, readonly=readonly):
        graph_dict[graph["iri"]] = graph
    return graph_dict


def convert_uri_to_filename(value: str, allow_unicode: bool = False) -> str:
    """Convert URI to unix friendly filename.

    Convert to ASCII if 'allow_unicode' is False. Convert spaces or repeated
    dashes to single dashes. Remove characters that aren't alphanumerics,
    underscores, or hyphens. Convert / to underscore. Convert to lowercase.
    Also strip leading and trailing whitespace, dashes, and underscores.
    """
    value = str(value)
    if allow_unicode:
        value = unicodedata.normalize("NFKC", value)
    else:
        value = unicodedata.normalize("NFKD", value).encode("ascii", "ignore").decode("ascii")
    value = re.sub(r"\.", "_", value.lower())
    value = re.sub(r"/", "_", value.lower())
    value = re.sub(r"[^\w\s-]", "", value.lower())
    return re.sub(r"[-\s]+", "-", value).strip("-_")


def struct_to_table(
    source: str | bool | float | list | dict, table: list | None = None, prefix: str = ""
) -> list:
    """Prepare flat key/value table from a deep structure.

    This function takes structure and creates a flat table out of it,
    key by key, value by value. For each level deeper it prefixes the
    father keys.

    Example input:  {'k1': '1', 'k2': {'k3': '3', 'k4': '4'}}
    Example output: [['k1', '1'], ['k2:k3', '3'], ['k2:k4', '4']]

    Args:
    ----
        source (any): The structure which is transformed into a flat table.
        table (list): The table where the key/value rows will be appended.
        prefix (str): A prefix which is used to indicate the level.

    Returns:
    -------
        The input table extended with rows from the input source.

    """
    if table is None:
        table = []
    if type(source) in (str, bool, int, float):
        table.append([prefix, source])
        return table
    if isinstance(source, dict):
        if len(prefix) != 0:
            prefix = prefix + "."
        for key in source:
            table = struct_to_table(source[key], table, prefix + key)
        return table
    if isinstance(source, list):
        for value in source:
            table = struct_to_table(value, table, prefix)
        return table
    return table


def split_task_id(task_id: str) -> tuple[str, str]:
    """Validate and split cmemc task ID.

    Args:
    ----
        task_id (str): The task ID in the workspace.

    Raises:
    ------
        ClickException: in case the task ID is not splittable

    """
    try:
        project_part = task_id.split(":")[0]
        task_part = task_id.split(":")[1]
    except IndexError as error:
        raise ClickException(f"{task_id} is not a valid task ID.") from error
    return project_part, task_part


def metric_get_labels(family: Metric, clean: bool = True) -> dict[str, list[str]]:
    """Get the labels of a metric family.

    clean: remove keys with only one dimension
    family: the metric family

    Returns: labels as dict
    """
    labels: dict[str, list[str]] = {}
    # build tree structure
    for sample in family.samples:
        for label in sample.labels:
            value = sample.labels[label]
            if label not in labels:
                labels[label] = []
            if value not in labels[label]:
                labels[label].append(value)
    if clean:
        labels = dict(filter(lambda elem: len(elem[1]) > 1, labels.items()))
    return labels


def check_or_select_project(app: "ApplicationContext", project_id: str | None = None) -> str:
    """Check for given project, select the first one if there is only one.

    Args:
    ----
        app (ApplicationContext): the click cli app context.
        project_id (str): The project ID.

    Raises:
    ------
        ClickException: if no projects available.
        ClickException: if more than one project is.

    Returns:
    -------
        Maybe project_id if there was no project_id before.

    """
    if project_id is not None:
        return project_id

    projects = get_projects()
    if len(projects) == 1:
        project_name = str(projects[0]["name"])
        app.echo_warning(
            "Missing project (--project) - since there is only one project, "
            f"this is selected: {project_name}"
        )
        return project_name

    if len(projects) == 0:
        raise ClickException(
            "There are no projects available. "
            "Please create a project with 'cmemc project create'."
        )

    # more than one project
    raise ClickException(
        "There is more than one project available so you need to "
        "specify the project with '--project'."
    )


@dataclass
class PublishedPackage:
    """Represents a published package from pypi.org."""

    name: str
    description: str
    published: str
    version: str


def get_published_packages() -> list[PublishedPackage]:
    """Get a scraped list of plugin packages scraped from pypi.org."""
    url = "https://download.eccenca.com/cmem-plugin-index/cmem-plugin-index.json"
    package_names = []
    packages = []
    for _ in requests.get(url, timeout=5).json():
        name = _["name"]
        if name == "cmem-plugin-base":
            continue
        if not name.startswith("cmem-plugin-"):
            continue
        if name not in package_names:
            package_names.append(name)
            packages.append(
                PublishedPackage(
                    name=name,
                    description=_["summary"],
                    published=_["latest_version_time"],
                    version=_["latest_version"],
                )
            )
    return packages


def convert_iri_to_qname(iri: str, default_ns: str) -> str:
    """Convert IRI to a QName based on the default namespace.

    If the IRI does not match the default namespace, the method returns the full IRI.
    """
    if iri.startswith(default_ns):
        return iri.replace(default_ns, ":")
    return iri


def convert_qname_to_iri(qname: str, default_ns: str) -> str:
    """Convert a QName to an IRI based on the default namespace.

    If the QName does not match the default namespace, the method returns the full IRI.
    """
    if qname.startswith(":"):
        return default_ns + qname[1:]

    return qname


def get_query_text(file_or_uri: str, required_projections: set) -> str:
    """Get query text for a file or graph catalog URI.

    Args:
    ----
        file_or_uri (str): The file path or URI to fetch the query from.
        required_projections (set): A set of required projections.

    Returns:
    -------
        str: The query text.

    Raises:
    ------
        ClickException: If the input is not a readable file or a query URI,
                    or if the query contains placeholders,
                    or if the query does not include the required projections.

    """
    sparql_query = QueryCatalog().get_query(file_or_uri)
    if sparql_query is None:
        raise ClickException(f"{file_or_uri} is neither a readable file nor a query URI.")

    if sparql_query.get_placeholder_keys():
        raise ClickException("Placeholder queries are not supported.")

    result = sparql_query.get_json_results()
    projected_vars = set(result["head"]["vars"])

    missing_projections = required_projections - projected_vars
    if missing_projections:
        missing = ", ".join(missing_projections)
        raise ClickException(f"Select query must include projections for: {missing}")
    txt: str = sparql_query.text
    return txt


def validate_zipfile(zipfile: str | pathlib.Path) -> bool:
    """Validate a zipfile."""
    zipfile = pathlib.Path(zipfile)
    try:
        ZipFile(zipfile).testzip()
    except BadZipFile:
        return False
    return True


def str_to_bool(value: str | bool) -> bool:
    """Convert common string representations of boolean values to True/False."""
    if isinstance(value, bool):  # If it's already a boolean, return it directly
        return value
    return str(value).strip().lower() in {"true", "yes", "1", "on"}


def is_enabled(params: dict[str, str], config: PureSectionConfigParser, key: str) -> bool:
    """Check if a parameter is enabled."""
    # Check in params dictionary (highest priority)
    if key in params and str_to_bool(params[key]):
        return True
    key = f"CMEMC_{key.upper()}"
    section = params.get("connection")
    defaults = config.defaults()
    if section is None:
        section = defaults.get("CMEMC_CONNECTION")

    # Check in the specified section
    if section and section in config and key in config[section]:
        return str_to_bool(config[section][key])

    # Check in the default section
    if key in defaults:
        return str_to_bool(config.defaults()[key])

    return False


def tuple_to_list(ctx: type["ApplicationContext"], param: Argument, value: tuple) -> list:  # noqa: ARG001
    """Get a list from a tuple

    Used as callback to have mutable values
    """
    return list(value)



"""dataset commands for cmem command line interface."""

import json
import re

import click
import requests.exceptions
from click import ClickException, UsageError
from cmem.cmempy.config import get_cmem_base_uri
from cmem.cmempy.workspace import get_task_plugin_description, get_task_plugins
from cmem.cmempy.workspace.projects.datasets.dataset import (
    create_dataset,
    delete_dataset,
    get_dataset,
    post_resource,
    update_dataset,
)
from cmem.cmempy.workspace.projects.resources.resource import (
    create_resource,
    get_resource_response,
    resource_exist,
)
from cmem.cmempy.workspace.search import list_items

from cmem_cmemc import completion
from cmem_cmemc.command import CmemcCommand
from cmem_cmemc.command_group import CmemcGroup
from cmem_cmemc.commands.resource import resource
from cmem_cmemc.completion import get_dataset_file_mapping
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.exceptions import CmemcError
from cmem_cmemc.parameter_types.path import ClickSmartPath
from cmem_cmemc.smart_path import SmartPath as Path
from cmem_cmemc.utils import check_or_select_project, struct_to_table

DATASET_FILTER_TYPES = sorted(["project", "regex", "tag", "type"])
DATASET_LIST_FILTER_HELP_TEXT = (
    "Filter datasets based on metadata. First parameter"
    f" can be one of the following values: {', '.join(DATASET_FILTER_TYPES)}."
    " The options for the second parameter depend on the first parameter."
)
DATASET_DELETE_FILTER_HELP_TEXT = (
    "Delete datasets based on metadata. First parameter --filter"
    f" CHOICE can be one of {DATASET_FILTER_TYPES!s}."
    " The second parameter is based on CHOICE."
)


def _get_dataset_tag_labels(dataset_: dict) -> list[str]:
    """Output a list of tag labels from a single dataset."""
    return [_["label"] for _ in dataset_["tags"]]


def _get_datasets_filtered(
    datasets: list[dict], filter_name: str, filter_value: str | int
) -> list[dict]:
    """Get dataset filtered according to filter name and value.

    Args:
    ----
        datasets: list of datasets to filter
        filter_name (str): one of DATASET_FILTER_TYPES
        filter_value (str|int): value according to filter

    Returns:
    -------
        list of filtered datasets from the get_query_status API call

    Raises:
    ------
        UsageError

    """
    if filter_name not in DATASET_FILTER_TYPES:
        raise UsageError(
            f"{filter_name} is an unknown filter name. " f"Use one of {DATASET_FILTER_TYPES}."
        )
    # filter by project ID
    if filter_name == "project":
        return [_ for _ in datasets if _["projectId"] == filter_value]
    # filter by regex on the label
    if filter_name == "regex":
        return [_ for _ in datasets if re.search(str(filter_value), _["label"])]
    # filter by dataset type
    if filter_name == "type":
        return [_ for _ in datasets if re.search(str(filter_value), _["pluginId"])]
    # filter by tag label
    if filter_name == "tag":
        return [_ for _ in datasets if filter_value in _get_dataset_tag_labels(_)]
    # default is unfiltered
    return datasets


def _validate_and_split_dataset_id(dataset_id: str) -> tuple[str, str]:
    """Validate and split cmemc dataset ID.

    Args:
    ----
        dataset_id (str): The cmemc dataset ID in the workspace.

    Raises:
    ------
        ClickException: in case the dataset ID is not splittable

    """
    try:
        project_part = dataset_id.split(":")[0]
        dataset_part = dataset_id.split(":")[1]
    except IndexError as error:
        raise ClickException(
            f"{dataset_id} is not a valid dataset ID. Use the "
            "'dataset list' command to get a list of existing datasets."
        ) from error
    return project_part, dataset_part


def _post_file_resource(
    app: ApplicationContext,
    project_id: str,
    dataset_id: str,
    local_file_name: str,
) -> None:
    """Upload a local file as a dataset resource to a project.

    Args:
    ----
        app: the click cli app context.
        project_id: The project ID in the workspace.
        dataset_id: The dataset ID in the workspace.
        local_file_name: The path to the local file name

    Raises:
    ------
        ValueError: if resource exists and no replace

    """
    app.echo_info(
        f"Upload {local_file_name} as a file resource of dataset "
        f"{dataset_id} to project {project_id} ... ",
        nl=False,
    )
    post_resource(
        project_id=project_id,
        dataset_id=dataset_id,
        file_resource=ClickSmartPath.open(local_file_name),
    )
    app.echo_success("done")


def _upload_file_resource(
    app: ApplicationContext,
    project_id: str,
    local_file_name: str,
    remote_file_name: str,
    replace: bool,
) -> None:
    """Upload a local file as a dataset resource to a project.

    Args:
    ----
        app: the click cli app context.
        project_id: The project ID in the workspace.
        local_file_name: The path to the local file name
        remote_file_name: The remote file name
        replace: Replace resource if needed.

    Raises:
    ------
        ValueError: if resource exists and no replace

    """
    exist = resource_exist(project_name=project_id, resource_name=remote_file_name)
    if exist and not replace:
        raise ClickException(
            f"A file resource with the name '{remote_file_name}' already "
            "exists in this project. \n"
            "Please rename the file or use the '--replace' "
            "parameter in order to overwrite the remote file."
        )
    if exist:
        app.echo_info(
            f"Replace content of {remote_file_name} with content from "
            f"{local_file_name} in project {project_id} ... ",
            nl=False,
        )
    else:
        app.echo_info(
            f"Upload {local_file_name} as a file resource "
            f"{remote_file_name} to project {project_id} ... ",
            nl=False,
        )
    create_resource(
        project_name=project_id,
        resource_name=remote_file_name,
        file_resource=ClickSmartPath.open(local_file_name),
        replace=replace,
    )
    app.echo_success("done")


def _get_metadata_out_of_parameter(parameter_dict: dict) -> dict:
    """Extract metadata keys out of the parameter dict.

    Args:
    ----
        parameter_dict (dict): the dictionary of given parameters.

    Returns:
    -------
        The dictionary of only the known metadata fields.

    """
    metadata_dict = {}
    if "label" in parameter_dict:
        metadata_dict["label"] = parameter_dict["label"]
    if "description" in parameter_dict:
        metadata_dict["description"] = parameter_dict["description"]
    return metadata_dict


def _get_read_only_out_of_parameter(parameter_dict: dict) -> bool:
    """Extract readonly key value out of the parameter dict.

    Args:
    ----
        parameter_dict (dict): the dictionary of given parameters.

    Returns:
    -------
        The value of read only field.

    """
    read_only = parameter_dict.get("readOnly", False)
    if read_only in ("true", True, "True"):
        return True
    if read_only in ("false", False, "False"):
        return False
    raise ClickException(f"readOnly parameter should be 'true' or 'false' - was {read_only!r}")


def _extend_parameter_with_metadata(
    app: ApplicationContext,
    parameter_dict: dict,
    dataset_type: str,
    dataset_file: str,
) -> dict:
    """Extend the parameter with label if needed.

    Args:
    ----
        app: the click cli app context.
        parameter_dict: The dictionary of given dataset parameters
        dataset_type: The dataset type ID
        dataset_file: The path of the local file

    Returns:
    -------
        An extended parameter dictionary (label + file)

    """
    if "label" not in parameter_dict:
        label = f"Unnamed {dataset_type} dataset"
        if dataset_file:
            label = Path(dataset_file).name
        if "file" in parameter_dict:
            label = parameter_dict["file"]
        app.echo_warning(
            "Missing dataset label (-p label xxx) - " f"this generated label will be used: {label}"
        )
        parameter_dict["label"] = label
    return parameter_dict


def _check_or_set_dataset_type(
    app: ApplicationContext, parameter_dict: dict, dataset_type: str, dataset_file: str
) -> str:
    """Check for missing dataset type.

    Args:
    ----
        app: the click cli app context.
        parameter_dict: The dictionary of given dataset parameters
        dataset_type: The dataset type ID.
        dataset_file: The path of the local file.

    Returns:
    -------
        A dataset type based the given file names.

    """
    source = Path(dataset_file).name if dataset_file else ""
    target = parameter_dict.get("file", "")
    suggestions = [
        (extension, info["type"]) for extension, info in get_dataset_file_mapping().items()
    ]
    if not dataset_type:
        for check, type_ in suggestions:
            if source.endswith(check) or target.endswith(check):
                dataset_type = type_
                break
        if not dataset_type:
            raise UsageError("Missing parameter. Please specify a dataset " "type with '--type'.")
        app.echo_warning(
            "Missing dataset type (--type) - based on the used file name, "
            f"this type is assumed: {dataset_type}"
        )
    return dataset_type


def _show_parameter_list(app: ApplicationContext, dataset_type: str) -> None:
    """Output the parameter list for a given dataset type.

    Args:
    ----
        app: the click cli app context.
        dataset_type: The type from which the parameters are listed.

    """
    plugin = get_task_plugin_description(dataset_type)
    properties = plugin["properties"]
    required_properties = plugin["required"]
    table = []
    for key in properties:
        if key in required_properties:
            parameter = key + " *"
            description = "(Required) " + properties[key]["description"]
        else:
            parameter = key
            description = properties[key]["description"]
        row = [
            parameter,
            description,
        ]
        table.append(row)

    table = completion.add_read_only_and_uri_property_parameters(table)

    # metadata always on top, then sorted by key
    table = sorted(table, key=lambda k: k[0].lower())
    table = completion.add_metadata_parameter(table)
    app.echo_info_table(table, headers=["Parameter", "Description"])


def _show_type_list(app: ApplicationContext) -> None:
    """Output the list of dataset types.

    Args:
    ----
        app: the click cli app context.

    """
    plugins = get_task_plugins()
    table = []
    for plugin_id in plugins:
        plugin = plugins[plugin_id]
        if plugin["taskType"] == "Dataset":
            id_ = plugin_id
            title = plugin["title"]
            description = plugin["description"].partition("\n")[0]
            row = [
                id_,
                f"{title}: {description}",
            ]
            table.append(row)
    app.echo_info_table(table, headers=["Dataset Type", "Description"], sort_column=1)


def _check_or_select_dataset_type(app: ApplicationContext, dataset_type: str) -> tuple[str, dict]:
    """Test type and return plugin.

    Args:
    ----
        app: the click cli app context.
        dataset_type: A dataset type

    Raises:
    ------
        CmemcError: If type is not known

    Returns:
    -------
        A tuple of dataset_type and corresponding plugin description (dict)

    """
    try:
        app.echo_debug(f"check type {dataset_type}")
        plugin = get_task_plugin_description(dataset_type)
    except requests.exceptions.HTTPError as error:
        raise CmemcError(app, f"Unknown dataset type: {dataset_type}.") from error
    else:
        return dataset_type, plugin


@click.command(cls=CmemcCommand, name="list")
@click.option(
    "--filter",
    "filter_",
    type=(str, str),
    multiple=True,
    shell_complete=completion.dataset_list_filter,
    help=DATASET_LIST_FILTER_HELP_TEXT,
)
@click.option(
    "--raw", is_flag=True, help="Outputs raw JSON objects of the dataset search API response."
)
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only dataset IDs and no labels or other metadata. "
    "This is useful for piping the IDs into other cmemc commands.",
)
@click.pass_obj
def list_command(
    app: ApplicationContext, filter_: tuple[tuple[str, str]], raw: bool, id_only: bool
) -> None:
    """List available datasets.

    Output and filter a list of available datasets. Each dataset is listed
    with its ID, type and label.
    """
    datasets = list_items(item_type="dataset")["results"]
    for _ in filter_:
        filter_type, filter_name = _
        datasets = _get_datasets_filtered(datasets, filter_type, filter_name)

    if raw:
        app.echo_info_json(datasets)
    elif id_only:
        for _ in datasets:
            app.echo_info(_["projectId"] + ":" + _["id"])
    else:
        table = []
        for _ in datasets:
            row = [
                _["projectId"] + ":" + _["id"],
                _["pluginId"],
                _["label"],
            ]
            table.append(row)
        app.echo_info_table(
            table,
            headers=["Dataset ID", "Type", "Label"],
            sort_column=2,
            empty_table_message="No datasets found. "
            "Use the `dataset create` command to create a new dataset.",
        )


@click.command(cls=CmemcCommand, name="delete")
@click.option(
    "-a",
    "--all",
    "all_",
    is_flag=True,
    help="Delete all datasets. " "This is a dangerous option, so use it with care.",
)
@click.option(
    "--project",
    "project_id",
    type=click.STRING,
    shell_complete=completion.project_ids,
    help="In combination with the '--all' flag, this option allows for "
    "deletion of all datasets of a certain project. The behaviour is "
    "similar to the 'dataset list --project' command.",
)
@click.option(
    "--filter",
    "filter_",
    type=(str, str),
    multiple=True,
    shell_complete=completion.dataset_list_filter,
    help=DATASET_DELETE_FILTER_HELP_TEXT,
)
@click.argument("dataset_ids", nargs=-1, type=click.STRING, shell_complete=completion.dataset_ids)
@click.pass_obj
def delete_command(
    app: ApplicationContext,
    project_id: str,
    all_: bool,
    filter_: tuple[tuple[str, str]],
    dataset_ids: tuple[str],
) -> None:
    """Delete datasets.

    This command deletes existing datasets in integration projects from
    Corporate Memory. The corresponding dataset resources will not be deleted.

    Warning: Datasets will be deleted without prompting.

    Note: Datasets can be listed by using the `dataset list` command.
    """
    if project_id:
        app.echo_warning(
            "Option '--project' is deprecated and will be removed. "
            "Please use '--filter project XXX' instead."
        )
    if dataset_ids == () and not all_ and not filter_:
        raise UsageError(
            "Either specify at least one dataset ID"
            " or use a --filter option,"
            " or use the --all option to delete all datasets."
        )

    if dataset_ids and (all_ or filter_):
        raise UsageError("Either specify a dataset ID OR" " use a --filter or the --all option.")

    if all_ or filter_:
        # in case --all or --filter is given, a list of datasets is fetched
        dataset_ids = []
        datasets = list_items(item_type="dataset", project=project_id)["results"]
        for _ in filter_:
            filter_type, filter_name = _
            datasets = _get_datasets_filtered(datasets, filter_type, filter_name)
        for _ in datasets:
            dataset_ids.append(_["projectId"] + ":" + _["id"])

    count = len(dataset_ids)
    current = 1
    for _ in dataset_ids:
        app.echo_info(f"Delete dataset {current}/{count}: {_} ... ", nl=False)
        project_part, dataset_part = _validate_and_split_dataset_id(_)
        app.echo_debug(f"Project ID is {project_part}, dataset ID is {dataset_part}")
        delete_dataset(project_part, dataset_part)
        app.echo_success("done")
        current = current + 1


@click.command(cls=CmemcCommand, name="download")
@click.argument("dataset_id", type=click.STRING, shell_complete=completion.dataset_ids)
@click.argument(
    "output_path",
    required=True,
    type=ClickSmartPath(allow_dash=True, dir_okay=False, writable=True),
)
@click.option(
    "--replace",
    is_flag=True,
    help="Replace existing files. This is a dangerous option, " "so use it with care!",
)
@click.pass_obj
def download_command(
    app: ApplicationContext, dataset_id: str, output_path: str, replace: bool
) -> None:
    """Download the resource file of a dataset.

    This command downloads the file resource of a dataset to your local
    file system or to standard out (`-`).
    Note that this is not possible for dataset types such as
    Knowledge Graph (`eccencaDataplatform`) or SQL endpoint (`sqlEndpoint`).

    Without providing an output path, the output file name will be the
    same as the remote file resource.

    Note: Datasets can be listed by using the `dataset list` command.
    """
    app.echo_debug(
        f"Dataset ID is {dataset_id}; "
        f"output path is {click.format_filename(output_path)}. "
        f"replace is {replace}."
    )
    project_part, dataset_part = _validate_and_split_dataset_id(dataset_id)
    project = get_dataset(project_part, dataset_part)
    try:
        file = project["data"]["parameters"]["file"]
    except KeyError as no_file_resource:
        raise CmemcError(
            app, f"The dataset {dataset_id} has no associated file resource."
        ) from no_file_resource
    if Path(output_path).exists() and replace is not True:
        raise UsageError(
            f"Target file {click.format_filename(output_path)} already "
            "exists. Use --replace in case you want to replace it."
        )
    with get_resource_response(project_part, file) as response:
        # if piping file to stdout, no info messages
        if output_path != "-":
            app.echo_info(
                f"Download resource {file} of dataset {dataset_id} to file "
                f"{click.format_filename(output_path)} ... ",
                nl=False,
            )
        with click.open_file(output_path, "wb") as resource_file:
            for chunk in response.iter_content(chunk_size=8192):
                resource_file.write(chunk)
            # if piping file to stdout, no info messages
        if output_path != "-":
            app.echo_success("done")


@click.command(cls=CmemcCommand, name="upload")
@click.argument("dataset_id", type=click.STRING, shell_complete=completion.dataset_ids)
@click.argument(
    "input_path",
    required=True,
    shell_complete=completion.dataset_files,
    type=ClickSmartPath(allow_dash=True, dir_okay=False, writable=True, remote_okay=True),
)
@click.pass_obj
def upload_command(app: ApplicationContext, dataset_id: str, input_path: str) -> None:
    """Upload a resource file to a dataset.

    This command uploads a file to a dataset.
    The content of the uploaded file replaces the remote file resource.
    The name of the remote file resource will not be changed.

    Warning: If the remote file resource is used in more than one dataset,
    all of these datasets are affected by this command.

    Warning: The content of the uploaded file is not tested, so uploading
    a JSON file to an XML dataset will result in errors.

    Note: Datasets can be listed by using the `dataset list` command.

    Example: cmemc dataset upload cmem:my-dataset new-file.csv
    """
    project_part, dataset_part = _validate_and_split_dataset_id(dataset_id)

    _post_file_resource(
        app=app,
        project_id=project_part,
        dataset_id=dataset_part,
        local_file_name=input_path,
    )


@click.command(cls=CmemcCommand, name="inspect")
@click.argument("dataset_id", type=click.STRING, shell_complete=completion.dataset_ids)
@click.option("--raw", is_flag=True, help="Outputs raw JSON.")
@click.pass_obj
def inspect_command(app: ApplicationContext, dataset_id: str, raw: bool) -> None:
    """Display metadata of a dataset.

    Note: Datasets can be listed by using the `dataset list` command.
    """
    app.echo_debug(f"Dataset ID is {dataset_id}")
    project_part, dataset_part = _validate_and_split_dataset_id(dataset_id)
    project = get_dataset(project_part, dataset_part)
    if raw:
        app.echo_info_json(project)
    else:
        table = struct_to_table(project)
        app.echo_info_table(table, headers=["Key", "Value"], sort_column=0)


@click.command(cls=CmemcCommand, name="create")
@click.argument(
    "DATASET_FILE",
    required=False,
    shell_complete=completion.dataset_files,
    type=ClickSmartPath(allow_dash=False, readable=True, exists=True, remote_okay=True),
)
@click.option(
    "--type",
    "-t",
    "dataset_type",
    multiple=False,
    type=click.STRING,
    shell_complete=completion.dataset_types,
    help="The dataset type of the dataset to create. Example types are 'csv',"
    "'json' and 'eccencaDataPlatform' (-> Knowledge Graph).",
)
@click.option(
    "--project",
    "project_id",
    type=click.STRING,
    shell_complete=completion.project_ids,
    help="The project, where you want to create the dataset in. If there is "
    "only one project in the workspace, this option can be omitted.",
)
@click.option(
    "--parameter",
    "-p",
    type=(str, str),
    shell_complete=completion.dataset_parameter,
    multiple=True,
    help="A set of key/value pairs. Each dataset type has different "
    "parameters (such as charset, arraySeparator, ignoreBadLines, ...). "
    "In order to get a list of possible parameter, use the"
    "'--help-parameter' option.",
)
@click.option(
    "--replace",
    is_flag=True,
    help="Replace remote file resources in case there " "already exists a file with the same name.",
)
@click.option(
    "--id",
    "dataset_id",
    type=click.STRING,
    help="The dataset ID of the dataset to create. "
    "The dataset ID will be automatically created in case it is not present.",
)
@click.option(
    "--help-types",
    is_flag=True,
    help="Lists all possible dataset types on given Corporate Memory instance. "
    "Note that this option already needs access to the instance.",
)
@click.option(
    "--help-parameter",
    is_flag=True,
    help="Lists all possible (optional and mandatory) parameter for a dataset "
    "type. Note that this option already needs access to the instance.",
)
@click.pass_obj
def create_command(  # noqa: PLR0913
    app: ApplicationContext,
    dataset_file: str,
    replace: bool,
    project_id: str,
    dataset_id: str,
    dataset_type: str,
    parameter: tuple[tuple[str, str]],
    help_parameter: bool,
    help_types: bool,
) -> None:
    """Create a dataset.

    Datasets are created in projects and can have associated file
    resources. Each dataset has a type (such as `csv`) and a list of
    parameters which can alter or specify the dataset behaviour.

    To get more information about available dataset types and associated
    parameters, use the --help-types and --help-parameter options.

    Example: cmemc dataset create --project my-project --type csv my-file.csv
    """
    if help_types:
        _show_type_list(app)
        return

    # transform the parameter list of tuple to a dictionary
    parameter_dict = dict(parameter)

    dataset_type = _check_or_set_dataset_type(
        app=app,
        parameter_dict=parameter_dict,
        dataset_type=dataset_type,
        dataset_file=dataset_file,
    )

    if help_parameter:
        _show_parameter_list(app, dataset_type=dataset_type)
        return

    dataset_type, plugin = _check_or_select_dataset_type(app, dataset_type)

    parameter_dict = _extend_parameter_with_metadata(
        app=app, parameter_dict=parameter_dict, dataset_type=dataset_type, dataset_file=dataset_file
    )

    project_id = check_or_select_project(app, project_id)

    # file required but not given
    if "file" in plugin["required"] and not dataset_file and "file" not in parameter_dict:
        raise UsageError(
            f"The dataset type {dataset_type} is file based, so you need "
            "to specify a file with the create command."
        )

    # file required and given
    # dataset_file = file path from the command line
    # parameter_dict["file"] = local name in DI
    if "file" in plugin["required"] and dataset_file:
        # add file parameter for the project if needed
        if "file" not in parameter_dict:
            parameter_dict["file"] = Path(dataset_file).name

        _upload_file_resource(
            app=app,
            project_id=project_id,
            local_file_name=dataset_file,
            remote_file_name=parameter_dict["file"],
            replace=replace,
        )

    # create dataset resource
    app.echo_info(f"Create a new dataset {project_id}:", nl=False)
    created_dataset = create_dataset(
        dataset_id=dataset_id,
        project_id=project_id,
        dataset_type=dataset_type,
        parameter=parameter_dict,
        metadata=_get_metadata_out_of_parameter(parameter_dict),
        read_only=_get_read_only_out_of_parameter(parameter_dict),
        uri_property=parameter_dict.get("uriProperty", ""),
    )
    returned_id = json.loads(created_dataset)["id"]
    app.echo_info(f"{returned_id} ... ", nl=False)
    app.echo_success("done")


@click.command(cls=CmemcCommand, name="update")
@click.argument(
    "dataset_id",
    type=click.STRING,
    required=True,
    shell_complete=completion.dataset_ids,
)
@click.option(
    "--parameter",
    "-p",
    type=(str, str),
    shell_complete=completion.dataset_parameter,
    multiple=True,
    help="A configuration parameter key/value pair. Each dataset type has different "
    "parameters (such as charset, arraySeparator, ignoreBadLines, ...). "
    "In order to get a list of possible parameter, use the"
    "'--help-parameter' option.",
)
@click.option(
    "--help-parameter",
    is_flag=True,
    help="Lists all possible (optional and mandatory) configuration parameter for"
    " a given dataset. Note that this option already needs access to the instance.",
)
@click.pass_obj
def update_command(
    app: ApplicationContext,
    dataset_id: str,
    parameter: tuple[tuple[str, str]],
    help_parameter: bool,
) -> None:
    """Update a dataset.

    With this command, you can update the configuration of an existing dataset.
    Similar to the `dataset create` command, you need to use configuration key/value
    pairs on the `--parameter` option.

    To get more information about the available configuration parameters on a dataset,
    use the `--help-parameter` option.

    Example: cmemc dataset update my-project:my-csv -p separator ";"
    """
    project_part, dataset_part = _validate_and_split_dataset_id(dataset_id)
    project = get_dataset(project_part, dataset_part)
    dataset_type = project["data"]["type"]

    if help_parameter:
        _show_parameter_list(app, dataset_type=dataset_type)
        return

    if not parameter:
        raise UsageError(
            "You need to use the `--parameter/-p` option at least once,"
            " in order to execute this command."
        )

    desc = get_task_plugin_description(dataset_type)
    possible_keys = ["label", "description", "readOnly"]
    possible_keys.extend(desc["properties"])
    possible_keys.extend(desc["required"])

    # transform the parameter list of tuple to a dictionary
    parameter_dict = {}
    for key, value in parameter:
        if key not in possible_keys:
            raise UsageError(
                f"Configuration key '{key}' is not valid for" f" the dataset type '{dataset_type}'."
            )
        parameter_dict[key] = value

    app.echo_info(f"Updating dataset {dataset_id} ... ", nl=False)
    update_dataset(
        dataset_id=dataset_part,
        project_id=project_part,
        parameters=parameter_dict,
        metadata=_get_metadata_out_of_parameter(parameter_dict),
        read_only=_get_read_only_out_of_parameter(parameter_dict),
        uri_property=parameter_dict.get("uriProperty", ""),
    )
    app.echo_success("done")


@click.command(cls=CmemcCommand, name="open")
@click.argument(
    "dataset_ids", nargs=-1, required=True, type=click.STRING, shell_complete=completion.dataset_ids
)
@click.pass_obj
def open_command(app: ApplicationContext, dataset_ids: tuple[str]) -> None:
    """Open datasets in the browser.

    With this command, you can open a dataset in the workspace in
    your browser.

    The command accepts multiple dataset IDs which results in
    opening multiple browser tabs.
    """
    dataset_urls = {}
    for _ in list_items(item_type="dataset")["results"]:
        dataset_id = _["projectId"] + ":" + _["id"]
        url = _["itemLinks"][0]["path"]
        dataset_urls[dataset_id] = url
    for _ in dataset_ids:
        if _ in dataset_urls:
            full_url = get_cmem_base_uri() + dataset_urls[_]
            app.echo_debug(f"Open {_}: {full_url}")
            click.launch(full_url)
        else:
            raise ClickException(f"Dataset '{_}' not found.")


@click.group(cls=CmemcGroup)
def dataset() -> CmemcGroup:  # type: ignore[empty-body]
    """List, create, delete, inspect, up-/download or open datasets.

    This command group allows for managing workspace datasets as well as
    dataset file resources. Datasets can be created and deleted.
    File resources can be uploaded and downloaded.
    Details of dataset parameters can be listed with inspect.

    Datasets are identified by a combined key of the PROJECT_ID and
    a DATASET_ID (e.g: `my-project:my-dataset`).

    Note: To get a list of existing datasets, execute the `dataset list`
    command or use tab-completion.
    """


dataset.add_command(list_command)
dataset.add_command(delete_command)
dataset.add_command(download_command)
dataset.add_command(upload_command)
dataset.add_command(inspect_command)
dataset.add_command(create_command)
dataset.add_command(open_command)
dataset.add_command(update_command)
dataset.add_command(resource)



"""Build (DataIntegration) variable commands for cmemc."""

import re

import click
from click import UsageError
from cmem.cmempy.workspace.projects.variables import (
    create_or_update_variable,
    delete_variable,
    get_all_variables,
    get_variable,
)

from cmem_cmemc import completion
from cmem_cmemc.command import CmemcCommand
from cmem_cmemc.command_group import CmemcGroup
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.utils import check_or_select_project, split_task_id

VARIABLES_FILTER_TYPES = ["project", "regex"]
VARIABLES_FILTER_TYPES_HIDDEN = ["ids"]
VARIABLES_FILTER_TEXT = (
    "Filter variables based on metadata. "
    f"First parameter CHOICE can be one of {VARIABLES_FILTER_TYPES!s}"
    ". The second parameter is based on CHOICE, e.g. a project "
    "ID or a regular expression string."
)


def _get_variables_filtered(
    variables: list[dict], filter_name: str, filter_value: str
) -> list[dict]:
    """Get variables but filtered according to name and value."""
    filter_types = VARIABLES_FILTER_TYPES + VARIABLES_FILTER_TYPES_HIDDEN
    # check for correct filter names (filter ids is used internally only)
    if filter_name not in filter_types:
        raise UsageError(
            f"{filter_name} is an unknown filter name. " f"Use one of {VARIABLES_FILTER_TYPES}."
        )
    # filter by ID list
    if filter_name == "ids":
        return [_ for _ in variables if _["id"] in filter_value]
    # filter by project
    if filter_name == "project":
        return [_ for _ in variables if _["project"] == filter_value]
    # filter by regex
    if filter_name == "regex":
        return [
            _
            for _ in variables
            if re.search(filter_value, _["id"])
            or re.search(filter_value, _["value"])
            or re.search(filter_value, _.get("description", ""))
            or re.search(filter_value, _.get("template", ""))
        ]
    # return unfiltered list
    return variables


@click.command(cls=CmemcCommand, name="list")
@click.option("--raw", is_flag=True, help="Outputs raw JSON.")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only variables names and no other metadata. "
    "This is useful for piping the IDs into other commands.",
)
@click.option(
    "--filter",
    "filters_",
    multiple=True,
    type=(str, str),
    shell_complete=completion.variable_list_filter,
    help=VARIABLES_FILTER_TEXT,
)
@click.pass_obj
def list_command(
    app: ApplicationContext, raw: bool, id_only: bool, filters_: tuple[tuple[str, str]]
) -> None:
    """List available project variables.

    Outputs a table or a list of project variables.
    """
    variables = get_all_variables()

    for _ in filters_:
        filter_name, filter_value = _
        variables = _get_variables_filtered(variables, filter_name, filter_value)
    if raw:
        app.echo_info_json(variables)
        return
    if id_only:
        for _ in sorted(_["id"] for _ in variables):
            app.echo_result(_)
        return
    # output a user table
    table = []
    headers = ["ID", "Value", "Template", "Description"]
    for _ in variables:
        row = [
            _["id"],
            _["value"],
            _.get("template", ""),
            _.get("description", ""),
        ]
        table.append(row)
    app.echo_info_table(
        table,
        headers=headers,
        sort_column=0,
        empty_table_message="No project variables found. "
        "Use the `project variable create` command to create a new project variable.",
    )


@click.command(cls=CmemcCommand, name="get")
@click.argument(
    "variable_id", required=True, type=click.STRING, shell_complete=completion.variable_ids
)
@click.option(
    "--key",
    type=click.Choice(["value", "template", "description"], case_sensitive=False),
    default="value",
    show_default=True,
    help="Specify the name of the value you want to get.",
)
@click.option("--raw", is_flag=True, help="Outputs raw json.")
@click.pass_obj
def get_command(app: ApplicationContext, variable_id: str, key: str, raw: bool) -> None:
    """Get the value or other data of a project variable.

    Use the `--key` option to specify which information you want to get.

    Note: Only the `value` key is always available on a project variable.
    Static value variables have no `template` key, and the `description` key
    is optional for both types of variables.
    """
    project_name, variable_name = split_task_id(variable_id)
    _ = get_variable(variable_name=variable_name, project_name=project_name)
    if raw:
        app.echo_info_json(_)
        return
    try:
        app.echo_info(_[key], nl=False)
    except KeyError as error:
        raise UsageError(f"Variable {variable_name} has no value of '{key}'.") from error


@click.command(cls=CmemcCommand, name="delete")
@click.argument(
    "variable_id", required=True, type=click.STRING, shell_complete=completion.variable_ids
)
@click.pass_obj
def delete_command(app: ApplicationContext, variable_id: str) -> None:
    """Delete a project variable.

    Note: You can not delete a variable which is used by another
    (template based) variable. In order to do so, delete the template based
    variable first.
    """
    project_name, variable_name = split_task_id(variable_id)
    app.echo_info(f"Delete variable {variable_name} from project {project_name} ... ", nl=False)
    delete_variable(variable_name=variable_name, project_name=project_name)
    app.echo_success("done")


# pylint: disable=too-many-arguments
@click.command(cls=CmemcCommand, name="create")
@click.argument(
    "variable_name",
    required=True,
    type=click.STRING,
)
@click.option("--value", type=click.STRING, help="The value of the new project variable.")
@click.option(
    "--template",
    type=click.STRING,
    help="The template of the new project variable. You can use Jinja template "
    "syntax, e.g. use '{{global.myVar}}' for accessing global variables, or "
    "'{{project.myVar}}' for accessing variables from the same project.",
)
@click.option(
    "--description", type=click.STRING, help="The optional description of the new project variable."
)
@click.option(
    "--project",
    "project_id",
    type=click.STRING,
    shell_complete=completion.project_ids,
    help="The project, where you want to create the variable in. If there is "
    "only one project in the workspace, this option can be omitted.",
)
@click.pass_obj
def create_command(  # noqa: PLR0913
    app: ApplicationContext,
    variable_name: str,
    value: str,
    template: str,
    description: str,
    project_id: str,
) -> None:
    """Create a new project variable.

    Variables need to be created with a value or a template (not both).
    In addition to that, a project ID and a name are mandatory.

    Example: cmemc project variable create my_var --project my_project --value abc

    Note: cmemc is currently not able to manage the order of the variables in a
    project. This means you have to create plain value variables in advance,
    before you can create template based variables, which access these values.
    """
    if value and template:
        raise UsageError("Either use '--value' or '--template' but not both.")
    if not value and not template:
        raise UsageError("Use '--value' or '--template' to create a new variable.")
    project_id = check_or_select_project(app, project_id)
    data = get_variable(project_name=project_id, variable_name=variable_name)
    if data:
        raise UsageError(f"Variable '{variable_name}' already exist in project '{project_id}'.")
    data = {"name": variable_name, "isSensitive": False, "scope": "project"}
    if value:
        data["value"] = value
    if template:
        data["template"] = template
    if description:
        data["description"] = description
    app.echo_info(f"Create variable {variable_name} in project {project_id} ... ", nl=False)
    create_or_update_variable(project_name=project_id, variable_name=variable_name, data=data)
    app.echo_success("done")


@click.command(cls=CmemcCommand, name="update")
@click.argument(
    "variable_id", required=True, type=click.STRING, shell_complete=completion.variable_ids
)
@click.option("--value", type=click.STRING, help="The new value of the project variable.")
@click.option(
    "--template",
    type=click.STRING,
    help="The new template of the project variable. You can use Jinja template "
    "syntax, e.g. use '{{global.myVar}}' for accessing global variables, or "
    "'{{project.myVar}}' for accessing variables from the same project.",
)
@click.option(
    "--description", type=click.STRING, help="The new description of the project variable."
)
@click.pass_obj
def update_command(
    app: ApplicationContext,
    variable_id: str,
    value: str,
    template: str,
    description: str,
) -> None:
    """Update data of an existing project variable.

    With this command you can update the value or the template, as well as the
    description of a project variable.

    Note: If you update the template of a static variable, it will be transformed
    to a template based variable. If you want to change the value of a template
    based variable, an error will be shown.
    """
    project_id, variable_name = split_task_id(variable_id)
    data = get_variable(project_name=project_id, variable_name=variable_name)
    if not data:
        raise UsageError(f"Variable '{variable_name}' does not exist in project '{project_id}'.")
    if value and template:
        raise UsageError(
            "Project variables are based on a static value or on a template, but not " "both."
        )
    if not value and not template and not description:
        raise UsageError(
            "Please specify what you want to update. "
            "Use at least one of the following options: "
            "'--value', '--template', '--description'."
        )
    if value:
        if data.get("template", None):
            raise UsageError("You can not change the value of a template based variable.")
        data["value"] = value
    if template:
        if not data.get("template", None):
            app.echo_warning(
                f"Variable '{variable_id}' will be converted from a "
                f"simple to a template based variable."
            )
        data["template"] = template
    if description:
        data["description"] = description
    app.echo_info(f"Update variable {variable_name} in project {project_id} ... ", nl=False)
    create_or_update_variable(project_name=project_id, variable_name=variable_name, data=data)
    app.echo_success("done")


@click.group(cls=CmemcGroup)
def variable() -> CmemcGroup:  # type: ignore[empty-body]
    """List, create, delete or get data from project variables.

    Project variables can be used in dataset and task parameters, and in the template
    transform operator.
    Variables are either based on a static value or based on a template.
    They may use templates that access globally configured
    variables or other preceding variables from the same project.

    Variables are identified by a VARIABLE_ID. To get a list of existing
    variables, execute the list command or use tab-completion.
    The VARIABLE_ID is a concatenation of a PROJECT_ID and a VARIABLE_NAME,
    such as `my-project:my-variable`.
    """


variable.add_command(list_command)
variable.add_command(get_command)
variable.add_command(delete_command)
variable.add_command(create_command)
variable.add_command(update_command)



"""Provides functionality for interacting with http/https paths"""

from __future__ import annotations

from typing import TYPE_CHECKING

from cmem_cmemc.smart_path.clients import StoragePath

if TYPE_CHECKING:
    from collections.abc import Generator


class HttpPath(StoragePath):
    """Client class for interacting with Amazon S3 storage.

    This class provides methods for working with http paths within the context
    of the `Path` application.
    """

    @property
    def suffix(self) -> str:
        """Return the suffix of the path."""
        raise NotImplementedError(f"suffix in {self.__class__} is not implemented yet.")

    @property
    def name(self) -> str:
        """Determine the name of the path."""
        return self.path.split("/")[-1]

    @property
    def parent(self) -> HttpPath:
        """The logical parent of the path."""
        raise NotImplementedError(f"parent in {self.__class__} is not implemented yet.")

    def is_dir(self) -> bool:
        """Determine if path is a directory or not."""
        return False

    def is_file(self) -> bool:
        """Determine if path is a file or not."""
        return not self.is_dir()

    def exists(self) -> bool:
        """Return the suffix of the path."""
        raise NotImplementedError(f"exists in {self.__class__} is not implemented yet.")

    def mkdir(self, parents: bool = False, exist_ok: bool = False) -> None:
        """Return the suffix of the path."""
        raise NotImplementedError(f"mkdir in {self.__class__} is not implemented yet.")

    def absolute(self) -> HttpPath:
        """Return an absolute version of this path"""
        raise NotImplementedError(f"absolute in {self.__class__} is not implemented yet.")

    def resolve(self) -> HttpPath:
        """Resolve the resolved path of the path."""
        raise NotImplementedError(f"resolve in {self.__class__} is not implemented yet.")

    def glob(self, pattern: str) -> Generator[StoragePath, StoragePath, StoragePath]:
        """Iterate over this subtree and yield all existing files"""
        raise NotImplementedError(f"glob in {self.__class__} is not implemented yet.")

    def __truediv__(self, key: str) -> HttpPath:
        """Return path with appending the key"""
        raise NotImplementedError(f"__truediv__ in {self.__class__} is not implemented yet.")



"""vocabularies commands for cmem command line interface."""

import io
from datetime import datetime, timezone
from re import match
from urllib.parse import urlparse

import click
from click import ClickException
from cmem.cmempy.config import get_cmem_base_uri
from cmem.cmempy.dp.proxy import graph as graph_api
from cmem.cmempy.dp.titles import resolve
from cmem.cmempy.queries import SparqlQuery
from cmem.cmempy.vocabularies import (
    get_global_vocabs_cache,
    get_vocabularies,
    install_vocabulary,
    uninstall_vocabulary,
)
from cmem.cmempy.workspace import reload_prefixes, update_global_vocabulary_cache
from rdflib import Graph
from rdflib.plugins.parsers.notation3 import BadSyntax
from six.moves.urllib.parse import quote

from cmem_cmemc import completion
from cmem_cmemc.command import CmemcCommand
from cmem_cmemc.command_group import CmemcGroup
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.parameter_types.path import ClickSmartPath

GET_ONTOLOGY_IRI_QUERY = """
PREFIX owl: <http://www.w3.org/2002/07/owl#>
SELECT DISTINCT ?iri
WHERE {
    ?iri a owl:Ontology;
}
"""

GET_PREFIX_DECLARATION = """
PREFIX owl: <http://www.w3.org/2002/07/owl#>
PREFIX vann: <http://purl.org/vocab/vann/>
SELECT DISTINCT ?prefix ?namespace
WHERE {{
    <{}> a owl:Ontology;
        vann:preferredNamespacePrefix ?prefix;
        vann:preferredNamespaceUri ?namespace.
}}
"""

INSERT_CATALOG_ENTRY = """
PREFIX owl: <http://www.w3.org/2002/07/owl#>
PREFIX voaf: <http://purl.org/vocommons/voaf#>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
PREFIX dct: <http://purl.org/dc/terms/>
PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
PREFIX vann: <http://purl.org/vocab/vann/>
PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
WITH <https://ns.eccenca.com/example/data/vocabs/>
INSERT {{
    <{iri}> a voaf:Vocabulary ;
        skos:prefLabel "{label}"{language} ;
        rdfs:label "{label}"{language} ;
        dct:description "{description}" ;
        vann:preferredNamespacePrefix "{prefix}" ;
        vann:preferredNamespaceUri "{namespace}" ;
        dct:modified "{date}"^^xsd:date .
}}
WHERE {{}}
"""


def _validate_vocabs_to_process(iris: tuple[str], filter_: str, all_flag: bool) -> list[str]:
    """Return a list of vocabulary IRTs which will be processed.

    list is without duplicates, and validated if they exist
    """
    if iris == () and not all_flag:
        raise click.UsageError(
            "Either specify at least one vocabulary IRI "
            "or use the --all option to process over all vocabularies."
        )
    all_vocabs = {_["iri"]: _ for _ in get_vocabularies()}
    if all_flag:
        # in case --all is given, all installable / installed vocabs are processed
        if filter_ == "installed":  # uninstall command
            return [_ for _ in all_vocabs if all_vocabs[_]["installed"]]
        # install command
        return [_ for _ in all_vocabs if not all_vocabs[_]["installed"]]

    vocabs_to_process = list(set(iris))  # avoid double removal / installation
    # test if one of the vocabs does not exist or is already installed / not installed
    for _ in vocabs_to_process:
        # uninstall command
        if filter_ == "installed" and (_ not in all_vocabs or not all_vocabs[_]["installed"]):
            raise click.UsageError(f"Vocabulary {_} not installed.")
        if filter_ == "installable":  # install command
            if _ not in all_vocabs:
                raise click.UsageError(f"Vocabulary {_} does not exist.")
            if all_vocabs[_]["installed"]:
                raise click.UsageError(f"Vocabulary {_} already installed.")
    return vocabs_to_process


def _validate_namespace(app: ApplicationContext, namespace: tuple[str | None, str | None]) -> None:
    """User input validation for the namespace."""
    prefix, uri = namespace
    if prefix is None or uri is None:
        raise ClickException("No namespace given.")

    if uri[-1] not in ("/", "#"):
        app.echo_warning(
            f"Warning: Namespace IRI '{uri}' does not end in"
            " hash (#) or slash (/). This is most likely not what you want."
        )
    parsed_url = urlparse(uri)
    app.echo_debug(str(parsed_url))
    if parsed_url.scheme not in ("http", "https", "urn"):
        raise ClickException(f"Namespace IRI '{uri}' is not a https(s) URL or an URN.")
    prefix_expression = r"^[a-z][a-z0-9]*$"
    if not match(prefix_expression, prefix):
        raise ClickException(
            "Prefix string does not match this regular" f" expression: {prefix_expression}"
        )


def _insert_catalog_entry(iri: str, prefix: str, namespace: str, label: str, language: str) -> None:
    """Insert a cmem vocabulary catalog entry.

    This executes an INSERT WHERE query to the vocabulary catalog graph in
    order to list the new vocabulary graph as vocab in the catalog.

    Args:
    ----
        iri (str): The IRI of the vocabulary graph.
        prefix (str): The prefix of the vocabulary.
        namespace (str): The namespace IRI of the vocabulary.
        label (str): The title of the vocabulary to add to the entry.
        language (str): Optional language tag of the title.

    Returns:
    -------
        None

    """
    language = "@" + str(language).strip() if "@" + str(language).strip() != "@" else ""

    if not label.startswith(prefix + ":"):
        label = prefix + ": " + label

    query_text = INSERT_CATALOG_ENTRY.format(
        iri=iri,
        prefix=prefix,
        namespace=namespace,
        date=datetime.now(tz=timezone.utc).date(),
        label=label,
        language=language,
        description="vocabulary imported with cmemc",
    )
    query = SparqlQuery(text=query_text, origin="cmemc")
    query.get_results()


def _get_vocabulary_metadata_from_file(
    file: io.BytesIO, namespace_given: bool = False
) -> dict[str, str]:
    """Get potential graph iri and prefix/namespace from a turtle file."""
    metadata = {"iri": "", "prefix": "", "namespace": ""}
    try:
        graph = Graph().parse(file, format="ttl")
    except BadSyntax as error:
        raise ClickException("File {file} could not be parsed as turtle.") from error

    ontology_iris = graph.query(GET_ONTOLOGY_IRI_QUERY)
    if len(ontology_iris) == 0:
        raise ClickException("There is no owl:Ontology resource described " "in the turtle file.")
    if len(ontology_iris) > 1:
        ontology_iris_str = [str(iri[0]) for iri in ontology_iris]  # type: ignore[index]
        raise ClickException(
            "There are more than one owl:Ontology resources described "
            f"in the turtle file: {ontology_iris_str}"
        )
    iri = str(next(iter(ontology_iris))[0])  # type: ignore[index]
    metadata["iri"] = iri
    vann_data = graph.query(GET_PREFIX_DECLARATION.format(iri))
    if not vann_data and not namespace_given:
        raise ClickException(
            "There is no namespace defined "
            f"for the ontology '{iri}'.\n"
            "Please add a prefix and namespace to the sources"
            "or use the --namespace option.\n"
            "Refer to the documentation at "
            "https://vocab.org/vann/ for more information."
        )
    if vann_data and namespace_given:
        raise ClickException(
            "There is already a namespace defined "
            f"in the file for the ontology '{iri}'.\n"
            "You can not use the --namespace option with this file."
        )
    if len(vann_data) > 1:
        raise ClickException(
            "There is more than one vann namespace defined " f"for the ontology: {iri}"
        )
    if not namespace_given:
        namespace = next(iter(vann_data))
        metadata["prefix"] = str(namespace[0])  # type: ignore[index]
        metadata["namespace"] = str(namespace[1])  # type: ignore[index]
    return metadata


def _transform_cache_to_table(cache_category: list[dict], table: list) -> list:
    """Transform a cache category dict to a tabulate table."""
    for item in cache_category:
        uri = item["genericInfo"]["uri"]
        try:
            label = item["genericInfo"]["label"]
        except KeyError:
            label = ""
        row = [uri, "class", label]
        table.append(row)
    return table


@click.command(cls=CmemcCommand, name="open")
@click.argument("iri", type=click.STRING, shell_complete=completion.installed_vocabularies)
@click.pass_obj
def open_command(app: ApplicationContext, iri: str) -> None:
    """Open / explore a vocabulary graph in the browser.

    Vocabularies are identified by their graph IRI.
    Installed vocabularies can be listed with the `vocabulary list` command.
    """
    explore_uri = get_cmem_base_uri() + "/explore?graph=" + quote(iri)
    click.launch(explore_uri)
    app.echo_debug(explore_uri)


@click.command(cls=CmemcCommand, name="list")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only vocabulary identifier (IRIs) and no labels or other "
    "metadata. This is useful for piping the ids into other cmemc "
    "commands.",
)
@click.option(
    "--filter",
    "filter_",
    type=click.Choice(["all", "installed", "installable"], case_sensitive=True),
    default="installed",
    show_default=True,
    help="Filter list based on status.",
)
@click.option("--raw", is_flag=True, help="Outputs raw JSON.")
@click.pass_obj
def list_command(app: ApplicationContext, id_only: bool, filter_: str, raw: bool) -> None:
    """Output a list of vocabularies.

    Vocabularies are graphs (see `graph` command group) which consists
    of class and property descriptions.
    """
    vocabs = get_vocabularies(filter_=filter_)
    if raw:
        app.echo_info_json(vocabs)
    elif id_only:
        for _ in vocabs:
            app.echo_info(_["iri"])
    else:
        table = []
        for _ in vocabs:
            iri = _["iri"]
            try:
                label = _["label"]["title"]
            except (KeyError, TypeError):
                label = _["vocabularyLabel"] if _["vocabularyLabel"] else "[no label given]"
            table.append((iri, label))
        app.echo_info_table(
            table,
            headers=["Vocabulary Graph IRI", "Label"],
            sort_column=1,
            empty_table_message="No installed vocabularies found. "
            "Use the `vocabulary install` command to install vocabulary from the catalog.",
        )


@click.command(cls=CmemcCommand, name="install")
@click.argument(
    "iris", nargs=-1, type=click.STRING, shell_complete=completion.installable_vocabularies
)
@click.option(
    "-a", "--all", "all_", is_flag=True, help="Install all vocabularies from the catalog."
)
@click.pass_obj
def install_command(app: ApplicationContext, iris: tuple[str], all_: bool) -> None:
    """Install one or more vocabularies from the catalog.

    Vocabularies are identified by their graph IRI.
    Installable vocabularies can be listed with the
    vocabulary list command.
    """
    vocabs_to_install = _validate_vocabs_to_process(iris=iris, filter_="installable", all_flag=all_)
    count: int = len(vocabs_to_install)
    for current, vocab in enumerate(vocabs_to_install, start=1):
        app.echo_info(f"Install vocabulary {current}/{count}: {vocab} ... ", nl=False)
        install_vocabulary(vocab)
        app.echo_success("done")


@click.command(cls=CmemcCommand, name="uninstall")
@click.argument(
    "iris", nargs=-1, type=click.STRING, shell_complete=completion.installed_vocabularies
)
@click.option("-a", "--all", "all_", is_flag=True, help="Uninstall all installed vocabularies.")
@click.pass_obj
def uninstall_command(app: ApplicationContext, iris: tuple[str], all_: bool) -> None:
    """Uninstall one or more vocabularies.

    Vocabularies are identified by their graph IRI.
    Already installed vocabularies can be listed with the
    vocabulary list command.
    """
    vocabs_to_uninstall = _validate_vocabs_to_process(iris=iris, filter_="installed", all_flag=all_)
    count: int = len(vocabs_to_uninstall)
    for current, vocab in enumerate(vocabs_to_uninstall, start=1):
        app.echo_info(f"Uninstall vocabulary {current}/{count}: {vocab} ... ", nl=False)
        uninstall_vocabulary(vocab)
        app.echo_success("done")


@click.command(cls=CmemcCommand, name="import")
@click.argument(
    "FILE",
    required=True,
    shell_complete=completion.triple_files,
    type=ClickSmartPath(allow_dash=True, readable=True, remote_okay=True),
)
@click.option(
    "--namespace",
    type=(str, str),
    default=(None, None),
    help="In case the imported vocabulary file does not include a preferred"
    " namespace prefix, you can manually add a namespace prefix"
    " with this option. Example: --namespace ex https://example.org/",
)
@click.option(
    "--replace", is_flag=True, help="Replace (overwrite) existing vocabulary, if present."
)
@click.pass_obj
def import_command(
    app: ApplicationContext, file: str, namespace: tuple[str | None, str | None], replace: bool
) -> None:
    """Import a turtle file as a vocabulary.

    With this command, you can import a local ontology file as a named graph
    and create a corresponding vocabulary catalog entry.

    The uploaded ontology file is analysed locally in order to discover the
    named graph and the prefix declaration. This requires an OWL ontology
    description which correctly uses the `vann:preferredNamespacePrefix` and
    `vann:preferredNamespaceUri` properties.
    """
    _buffer = io.BytesIO()
    with ClickSmartPath.open(file) as file_handle:
        _buffer.write(file_handle.read())
    _buffer.seek(0)

    # fetch metadata
    if namespace != (None, None):
        _validate_namespace(app, namespace)
        meta_data = _get_vocabulary_metadata_from_file(_buffer, True)
        meta_data["prefix"] = namespace[0]  # type: ignore[assignment]
        meta_data["namespace"] = namespace[1]  # type: ignore[assignment]
    else:
        meta_data = _get_vocabulary_metadata_from_file(_buffer, False)
    iri = meta_data["iri"]

    success_message = "done"
    if iri in [_["iri"] for _ in graph_api.get_graphs_list()]:
        if replace:
            success_message = "replaced"
        else:
            raise ClickException(f"Proposed graph {iri} does already exist.")
    app.echo_info(f"Import {file} as vocabulary to {iri} ... ", nl=False)
    # upload graph
    _buffer.seek(0)
    graph_api.post_streamed(iri, _buffer, replace=True)

    # resolve label
    resolved_label_object: dict = resolve([iri], graph=iri)[iri]
    app.echo_debug(str(resolved_label_object))
    label = resolved_label_object.get("title", iri)
    language = resolved_label_object.get("lang", "")

    # insert catalog entry
    _insert_catalog_entry(
        iri=iri,
        prefix=meta_data["prefix"],
        namespace=meta_data["namespace"],
        label=label,
        language=language,
    )
    # reload DI prefix
    reload_prefixes()
    # update cache
    update_global_vocabulary_cache(iri)
    app.echo_success(success_message)


@click.command(cls=CmemcCommand, name="update")
@click.argument(
    "iris", nargs=-1, type=click.STRING, shell_complete=completion.installed_vocabularies
)
@click.option(
    "-a", "--all", "all_", is_flag=True, help="Update cache for all installed vocabularies."
)
@click.pass_obj
def cache_update_command(app: ApplicationContext, iris: tuple[str], all_: bool) -> None:
    """Reload / updates the data integration cache for a vocabulary."""
    vocab_caches_to_update = _validate_vocabs_to_process(
        iris=iris, filter_="installed", all_flag=all_
    )
    count: int = len(vocab_caches_to_update)
    for current, vocab in enumerate(vocab_caches_to_update, start=1):
        app.echo_info(f"Update cache {current}/{count}: {vocab} ... ", nl=False)
        update_global_vocabulary_cache(vocab)
        app.echo_success("done")


@click.command(cls=CmemcCommand, name="list")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only vocabulary term identifier (IRIs) and no labels or other "
    "metadata. This is useful for piping the ids into other cmemc "
    "commands.",
)
@click.option("--raw", is_flag=True, help="Outputs raw JSON.")
@click.pass_obj
def cache_list_command(app: ApplicationContext, id_only: bool, raw: bool) -> None:
    """Output the content of the global vocabulary cache."""
    cache_ = get_global_vocabs_cache()
    if raw:
        app.echo_info_json(cache_)
    elif id_only:
        for vocab in cache_["vocabularies"]:
            for class_ in vocab["classes"]:
                app.echo_info(class_["genericInfo"]["uri"])
            for property_ in vocab["properties"]:
                app.echo_info(property_["genericInfo"]["uri"])
    else:
        table: list[list] = []
        for vocab in cache_["vocabularies"]:
            table = _transform_cache_to_table(vocab["classes"], table)
            table = _transform_cache_to_table(vocab["properties"], table)
        app.echo_info_table(
            table,
            headers=["IRI", "Type", "Label"],
            sort_column=0,
            empty_table_message="No cache entries found. "
            "Use the `vocabulary install` command to install a vocabulary.",
        )


@click.group(cls=CmemcGroup)
def cache() -> CmemcGroup:  # type: ignore[empty-body]
    """List und update the vocabulary cache."""


cache.add_command(cache_update_command)
cache.add_command(cache_list_command)


@click.group(cls=CmemcGroup)
def vocabulary() -> CmemcGroup:  # type: ignore[empty-body]
    """List, (un-)install, import or open vocabs / manage cache."""


vocabulary.add_command(open_command)
vocabulary.add_command(list_command)
vocabulary.add_command(install_command)
vocabulary.add_command(uninstall_command)
vocabulary.add_command(import_command)
vocabulary.add_command(cache)
